#
# Simple species module. 
#
# TODO: replace with a real package
#

import scipy.constants

mec2 = scipy.constants.value('electron mass energy equivalent in MeV')*1e6
mpc2 = scipy.constants.value('proton mass energy equivalent in MeV')*1e6

e_charge = scipy.constants.e
c_light = scipy.constants.c

CHARGE_OF = {'electron': -e_charge,
            'positron': e_charge,
            'proton': e_charge}

CHARGE_STATE = {
    'electron': -1,
    'positron': 1,
    'proton': 1}


MASS_OF = {'electron': mec2,
           'positron': mec2,
           'proton': mpc2}




# Functions

def mass_of(species):
    if species in MASS_OF:
        return MASS_OF[species]
    
    raise ValueError(f'Species not available: {species}')
    
    
def charge_of(species):
    if species in CHARGE_OF:
        return CHARGE_OF[species]
    
    raise ValueError(f'Species not available: {species}')
    
def charge_state(species):
    if species in CHARGE_STATE:
        return CHARGE_STATE[species]
    
    raise ValueError(f'Species not available: {species}')    
    
    
    


import numpy as np

from .units import unit, pg_units
from .readers import component_from_alias, load_field_attrs
from .tools import fstr, encode_attrs


def pmd_init(h5, basePath='/data/%T/', particlesPath='./' ):
    """
    Root attribute initialization.
    
    h5 should be the root of the file.
    """
    d = {
        'basePath':basePath,
        'dataType':'openPMD',
        'openPMD':'2.0.0',
        'openPMDextension':'BeamPhysics;SpeciesType',
        'particlesPath':particlesPath    
    }
    for k,v in d.items():
        h5.attrs[k] = fstr(v)
        
    
def pmd_field_init(h5, externalFieldPath='/ExternalFieldPath/%T/'):
    """
    Root attribute initialization for an openPMD-beamphysics External Field Mesh
    
    h5 should be the root of the file.
    """
    d = {
        'dataType':'openPMD',
        'openPMD':'2.0.0',
        'openPMDextension':'BeamPhysics',
        'externalFieldPath': externalFieldPath
    }
    for k,v in d.items():
        h5.attrs[k] = fstr(v)    
        
    
        
def write_pmd_bunch(h5, data, name=None):
    """
    Data is a dict with:
        np.array: 'x', 'px', 'y', 'py', 'z', 'pz', 't', 'status', 'weight'
        str: 'species'
        int: n_particle

    Optional data:
        np.array: 'id'
        
    See inverse routine:
        .particles.load_bunch_data
    
    """
    if name:
        g = h5.create_group(name)
    else:
        g = h5
    
    # Attributes
    g.attrs['speciesType'] = fstr( data['species'] )
    g.attrs['numParticles'] = data['n_particle']
    g.attrs['totalCharge'] = data['charge']
    g.attrs['chargeUnitSI'] = 1.0
    
    # Required Datasets
    for key in ['x', 'px', 'y', 'py', 'z', 'pz', 't', 'status', 'weight']:
        # Get full name, write data
        g2_name = component_from_alias[key]
        
        # Units
        u = pg_units(key)        
        
        # Write
        g2 = write_component_data(g, g2_name, data[key], unit=u)
    
        
    # Optional id. This does not have any units.
    if 'id' in data:
        g['id'] = data['id']

        
def write_pmd_field(h5, data, name=None):
    """
    Data is a dict with:
        attrs: flat dict of attributes. 
        components: flat dict of components
        
    See inverse routine:
        .readers.load_field_data
    
    """
    if name:
        g = h5.create_group(name)
    else:
        g = h5
    
    # Validate attrs
    attrs, other = load_field_attrs(data['attrs'])

    # Encode and write required and optional
    attrs = encode_attrs(attrs)
    for k, v in attrs.items():
        g.attrs[k] = v
    
    # All other attributes (don't change them)
    for k, v in other.items():
        g.attrs[k] = v
    
    # write components (datasets)
    for key, val in data['components'].items():
        
        # Units
        u = pg_units(key)   
        
        # Ensure complex
        val = val.astype(np.complex)
        
        # Write
        g2 = write_component_data(g, key, val, unit=u)            
    
    
def write_component_data(h5, name, data, unit=None): 
    """
    Writes data to a dataset h5[name]
    
    If data is a constant array, a group is created with the constant value and shape
    
    If unit is given, this will be used 
    
    """
    # Check for constant component
    dat0 = data[0]
    if np.all(data == dat0):
        g = h5.create_group(name)
        g.attrs['value'] = dat0
        g.attrs['shape'] = data.shape
    else:
        h5[name] = data
        g = h5[name]
        if len(data.shape) > 1:
            g.attrs['gridDataOrder'] = fstr('C') # C order for numpy/h5py
    
    if unit:
        g.attrs['unitSI'] = unit.unitSI
        g.attrs['unitDimension'] = unit.unitDimension
        g.attrs['unitSymbol'] = fstr(unit.unitSymbol)
    
    return g         






"""
Simple units functionality for the openPMD beamphysics records.

For more advanced units, use a package like Pint:
    https://pint.readthedocs.io/


"""
import scipy.constants

mec2 = scipy.constants.value('electron mass energy equivalent in MeV')*1e6
mpc2 = scipy.constants.value('proton mass energy equivalent in MeV')*1e6
c_light = scipy.constants.c
e_charge = scipy.constants.e
mu_0 = scipy.constants.mu_0 # Note that this is no longer 4pi*10^-7 !
 
import numpy as np


class pmd_unit:
    """
    
    Params
    ------
    
    unitSymbol: Native units name
    unitSI:     Conversion factor to the the correspontign SI unit
    unitDimension: SI Base Exponents

    Base unit dimensions are defined as:    
       Base dimension  | exponents.       | SI unit
       ---------------- -----------------   -------
       length          : (1,0,0,0,0,0,0)     m
       mass            : (0,1,0,0,0,0,0)     kg
       time            : (0,0,1,0,0,0,0)     s
       current         : (0,0,0,1,0,0,0)     A
       temperture      : (0,0,0,0,1,0,0)     K
       mol             : (0,0,0,0,0,1,0)     mol
       luminous        : (0,0,0,0,0,0,1)     cd 

    Example:
        pmd_unit('eV', 1.602176634e-19, (2, 1, -2, 0, 0, 0, 0))
        defines that an eV is 1.602176634e-19 of base units m^2 kg/s^2, which is a Joule (J)
    
    If unitSI=0 (default), init with a known symbol:
        pmd_unit('T')
    returns:
        pmd_unit('T', 1, (0, 1, -2, -1, 0, 0, 0))
        
        
    Simple equalities are provided:
        u1 == u2
    Returns True if the params are all the same. 
    
    """
    def __init__(self, unitSymbol='', unitSI=0, unitDimension = (0,0,0,0,0,0,0)):
        
        # Allow to return an internally known unit
        if unitSI==0:
            if unitSymbol in known_unit:
                # Copy internals
                u =  known_unit[unitSymbol]
                unitSI= u.unitSI  
                unitDimension = u.unitDimension
            else:
                raise ValueError(f'unknown unitSymbol: {unitSymbol}')
                
        self._unitSymbol = unitSymbol       
        self._unitSI = unitSI
        if isinstance(unitDimension, str):
             self._unitDimension = DIMENSION[unitDimension]
        else:
            self._unitDimension = unitDimension
            
    @property 
    def unitSymbol(self):
        return self._unitSymbol
    @property
    def unitSI(self):
        return self._unitSI
    @property
    def unitDimension(self):
        return self._unitDimension 

    def __mul__(self, other):
        return multiply_units(self, other)
    
    def __truediv__(self, other):
        return divide_units(self, other)
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)    
    
    
    def __str__(self):
        return self.unitSymbol
    
    def __repr__(self):
        return  f"pmd_unit('{self.unitSymbol}', {self.unitSI}, {self.unitDimension})"
  

def is_identity(u):
    """Checks if the unit is equivalent to 1"""
    return u.unitSI == 1 and u.unitDimension == (0,0,0,0,0,0,0)

def multiply_units(u1, u2):
    """
    Multiplies two pmd_unit symbols
    """

    if is_identity(u1):
        return u2
    if is_identity(u2):
        return u1    
    
    s1 = u1.unitSymbol
    s2 = u2.unitSymbol
    if s1==s2:
        symbol = f'({s1})^2'
    else:
        symbol = s1+'*'+s2
    d1 = u1.unitDimension
    d2 = u2.unitDimension
    dim=tuple(sum(x) for x in zip(d1, d2))
    unitSI = u1.unitSI * u2.unitSI
    
    return pmd_unit(unitSymbol=symbol, unitSI=unitSI, unitDimension=dim)     

def divide_units(u1, u2):
    """
    Divides two pmd_unit symbols : u1/u2
    """

    if is_identity(u2):
        return u1
    
    s1 = u1.unitSymbol
    s2 = u2.unitSymbol
    if s1==s2:
        symbol =  '1'
    else:
        symbol = s1+'/'+s2
    d1 = u1.unitDimension
    d2 = u2.unitDimension
    dim=tuple(a-b for a,b in zip(d1, d2))
    unitSI = u1.unitSI / u2.unitSI
    
    return pmd_unit(unitSymbol=symbol, unitSI=unitSI, unitDimension=dim)   
        
def sqrt_unit(u):
    """
    Returns the sqrt of a unit
    """
    u.unitDimension
    
    symbol = u.unitSymbol
    if symbol not in ['', '1']:
        symbol = f'sqrt({symbol})'
    
    unitSI = np.sqrt(u.unitSI)
    dim = tuple( x/2 for x in u.unitDimension)
    
    return pmd_unit(unitSymbol=symbol, unitSI=unitSI, unitDimension=dim)   
        
        
DIMENSION = { 
    '1'              : (0,0,0,0,0,0,0),
     # Base units
    'length'         : (1,0,0,0,0,0,0),
    'mass'           : (0,1,0,0,0,0,0),
    'time'           : (0,0,1,0,0,0,0),
    'current'        : (0,0,0,1,0,0,0),
    'temperture'     : (0,0,0,0,1,0,0),
    'mol'            : (0,0,0,0,0,1,0),
    'luminous'       : (0,0,0,0,0,0,1),
    #
    'charge'         : (0,0,1,1,0,0,0),
    'electric_field'  : (1,1,-3,-1,0,0,0),
    'electric_potential' : (1,2,-3,-1,0,0,0),
    'magnetic_field' : (0,1,-2,-1,0,0,0),    
    'velocity'       : (1,0,-1,0,0,0,0),
    'energy'         : (2,1,-2,0,0,0,0),
    'momentum'       : (1,1,-1,0,0,0,0)
}
# Inverse
DIMENSION_NAME = {v: k for k, v in DIMENSION.items()}

def dimension(name):
    if name in DIMENSION:
        return DIMENSION[name]
    else:
        return None

def dimension_name(dim_array):
    return DIMENSION_NAME[tuple(dim_array)]

SI_symbol = {
    '1'              : '1',
    'length'         : 'm',
    'mass'           : 'kg',
    'time'           : 's',
    'current'        : 'A',
    'temperture'     : 'K',
    'mol'            : 'mol',
    'luminous'       : 'cd',
    'charge'         : 'C',
    'electric_field' : 'V/m',
    'electric_potential' : 'V',
    'velocity'       : 'm/s',
    'energy'         : 'J',
    'momentum'       : 'kg*m/s',
    'magnetic_field' : 'T'
}
# Inverse
SI_name = {v: k for k, v in SI_symbol.items()}



known_unit = { 
    '1'          : pmd_unit('', 1, '1'),
    'rad'        : pmd_unit('', 1, '1'),
    'm'          : pmd_unit('m', 1, 'length'),
    'kg'         : pmd_unit('kg', 1, 'mass'),
    'g'          : pmd_unit('g', .001, 'mass'),
    's'          : pmd_unit('s', 1, 'time'),
    'A'          : pmd_unit('A', 1, 'current'),
    'K'          : pmd_unit('K', 1, 'temperture'),
    'mol'        : pmd_unit('mol', 1, 'mol'),
    'cd'         : pmd_unit('cd', 1, 'luminous'),
    'C'          : pmd_unit('C', 1, 'charge'),
    'charge_num' : pmd_unit('charge #', 1, 'charge'),
    'V/m'        : pmd_unit('V/m', 1, 'electric_field'),
    'V'          : pmd_unit('V', 1, 'electric_potential'),
    'c_light'    : pmd_unit('vel/c', c_light, 'velocity'),
    'm/s'        : pmd_unit('m/s', 1, 'velocity'),
    'eV'         : pmd_unit('eV', e_charge, 'energy'),
    'J'          : pmd_unit('J', 1, 'energy'),
    'eV/c'       : pmd_unit('eV/c', e_charge/c_light, 'momentum'),
    'T'          : pmd_unit('T', 1, 'magnetic_field')
    } 

def unit(symbol):
    """
    Returns a pmd_unit from a known symbol.
    
    * is allowed between two known symbols: 
    """
    if symbol in known_unit:
        return known_unit[symbol]
    
    if '*' in symbol:
        subunits = [known_unit[s] for s in symbol.split('*')]
        # Require these to be in known units
        assert len(subunits) == 2, 'TODO: more complicated units'
        return multiply_units(subunits[0], subunits[1])
    
    raise ValueError(f'Unknown unit symbol: {symbol}')
        
    







# Dicts for prefixes
PREFIX_FACTOR = {
    'yocto-' :1e-24,
    'zepto-' :1e-21,
    'atto-'  :1e-18,
    'femto-' :1e-15,
    'pico-'  :1e-12,
    'nano-'  :1e-9 ,
    'micro-' :1e-6,
    'milli-' :1e-3 ,
    'centi-' :1e-2 ,
    'deci-'  :1e-1,
    'deca-'  :1e+1,
    'hecto-' :1e2  ,
    'kilo-'  :1e3  ,
    'mega-'  :1e6  ,
    'giga-'  :1e9  ,
    'tera-'  :1e12 ,
    'peta-'  :1e15 ,
    'exa-'   :1e18 ,
    'zetta-' :1e21 ,
    'yotta-' :1e24
}
# Inverse
PREFIX = dict( (v,k) for k,v in PREFIX_FACTOR.items())

SHORT_PREFIX_FACTOR = {
    'y'  :1e-24,
    'z'  :1e-21,
    'a'  :1e-18,
    'f'  :1e-15,
    'p'  :1e-12,
    'n'  :1e-9 ,
    'µ'  :1e-6,
    'm'  :1e-3 ,
    'c'  :1e-2 ,
    'd'  :1e-1,
    ''   : 1,
    'da' :1e+1,
    'h'  :1e2  ,
    'k'  :1e3  ,
    'M'  :1e6  ,
    'G'  :1e9  ,
    'T'  :1e12 ,
    'P'  :1e15 ,
    'E'  :1e18 ,
    'Z'  :1e21 ,
    'Y'  :1e24
}
# Inverse
SHORT_PREFIX = dict( (v,k) for k,v in SHORT_PREFIX_FACTOR.items())




# Nice scaling

def nice_scale_prefix(scale):
    """
    Returns a nice factor and a SI prefix string 
    
    Example:
        scale = 2e-10
        
        f, u = nice_scale_prefix(scale)
        
        
    """
    
    if scale == 0:
        return 1, ''
    
    p10 = np.log10(abs(scale))

    if p10 <-1.5 or p10 > 2:
        f = 10**(p10 //3 *3)
    else:
        f = 1
    
    return f, SHORT_PREFIX[f]

def nice_array(a):
    """
    Returns a scaled array, the scaling, and a unit prefix
    
    Example: 
        nice_array( np.array([2e-10, 3e-10]) )
    Returns:
        (array([200., 300.]), 1e-12, 'p')
    
    """
    
    if np.isscalar(a):
        x = a
    elif len(a) == 1:
        x = a[0]
    else:
        a = np.array(a)
        x = a.ptp()
     
    fac, prefix = nice_scale_prefix( x )
    
    return a/fac, fac,  prefix




# -------------------------
# Units for ParticleGroup

PARTICLEGROUP_UNITS = {}
for k in ['status']:
    PARTICLEGROUP_UNITS[k] = unit('1')
for k in ['t']:
    PARTICLEGROUP_UNITS[k] = unit('s')
for k in ['energy', 'kinetic_energy', 'mass', 'higher_order_energy_spread', 'higher_order_energy']:
    PARTICLEGROUP_UNITS[k] = unit('eV')
for k in ['px', 'py', 'pz', 'p', 'pr', 'ptheta']:
    PARTICLEGROUP_UNITS[k] = unit('eV/c')
for k in ['x', 'y', 'z', 'r', 'Jx', 'Jy']:
    PARTICLEGROUP_UNITS[k] = unit('m')
for k in ['beta', 'beta_x', 'beta_y', 'beta_z', 'gamma', 'theta']:    
    PARTICLEGROUP_UNITS[k] = unit('1')
for k in ['charge', 'species_charge', 'weight']:
    PARTICLEGROUP_UNITS[k] = unit('C')
for k in ['average_current']:
    PARTICLEGROUP_UNITS[k] = unit('A')
for k in ['norm_emit_x', 'norm_emit_y']:
    PARTICLEGROUP_UNITS[k] = unit('m')
for k in ['norm_emit_4d']:
    PARTICLEGROUP_UNITS[k] = multiply_units(unit('m'), unit('m'))
for k in ['Lz']:
    PARTICLEGROUP_UNITS[k] = multiply_units(unit('m'), unit('eV/c'))    
for k in ['xp', 'yp']:
    PARTICLEGROUP_UNITS[k] = unit('rad')
for k in ['x_bar', 'px_bar', 'y_bar', 'py_bar']:
    PARTICLEGROUP_UNITS[k] = sqrt_unit(unit('m'))

def pg_units(key):
    """
    Returns a str representing the units of any attribute
    """
    for prefix in ['sigma_', 'mean_', 'min_', 'max_', 'ptp_', 'delta_']:
        if key.startswith(prefix):
            nkey = key[len(prefix):]
            return PARTICLEGROUP_UNITS[nkey]
    
    if key.startswith('cov_'):
        subkeys = key.strip('cov_').split('__')
        unit0 = PARTICLEGROUP_UNITS[subkeys[0]] 

        unit1 = PARTICLEGROUP_UNITS[subkeys[1]] 
        
        return multiply_units(unit0, unit1)
   
    # Fields
    if key.startswith('electricField'):
        return unit('V/m')  
    if key.startswith('magneticField'):
        return unit('T')
  
    
    
    return PARTICLEGROUP_UNITS[key]    
    
    
# -------------------------
# h5 tools

def write_unit_h5(h5, u):
    """
    Writes an pmd_unit to an h5 handle
    """

    h5.attrs['unitSI'] = u.unitSI
    h5.attrs['unitDimension'] = u.unitDimension
    h5.attrs['unitSymbol'] = u.unitSymbol
    
def read_unit_h5(h5):
    """
    Reads unit data from an h5 handle and returns a pmd_unit object
    """
    a = h5.attrs
    
    unitSI = a['unitSI']
    unitDimension = tuple(a['unitDimension'])
    if 'unitSymbol' not in a:
        unitSymbol = 'unknown'
    else:
        unitSymbol=a['unitSymbol']
        
    return pmd_unit(unitSymbol=unitSymbol, unitSI=unitSI, unitDimension=unitDimension)


    



def read_dataset_and_unit_h5(h5, expected_unit=None, convert=True):
    """
    Reads a dataset that has openPMD unit attributes.
    
    expected_unit can be a pmd_unit object, or a known unit str. Examples: 'kg', 'J', 'eV'
    
    If expected_unit is given, will check that the units are compatible.
    
    If convert, the data will be returned with the expected_units. 
    
    
    Returns a tuple:
        np.array, pmd_unit
        
    """
    
    # Read the unit that is there. 
    u = read_unit_h5(h5)
    
    # Simple case
    if not expected_unit:
        return np.array(h5), u
    
    if isinstance(expected_unit, str):
        # Try to get unit
        expected_unit = unit(expected_unit)
        
    # Check dimensions
    du = divide_units(u, expected_unit)

    assert du.unitDimension ==  (0,0,0,0,0,0,0), 'incompatible units'
    
    if convert:
        fac = du.unitSI
        return fac*np.array(h5), expected_unit
    else:
        return np.array(h5), u

    
def write_dataset_and_unit_h5(h5, name, data, unit=None):
    """
    Writes data and pmd_unit to h5[name]
    
    See: read_dataset_and_unit_h5
    """
    h5[name] = data
    
    if unit:
        write_unit_h5(h5[name], unit)
    
    

    
    


    




import scipy.io as sio
import numpy as np

def lucretia_to_data(filename, ele_name='BEGINNING', t_ref=0, exclude_dead_particles=True, verbose=False):

    """
    Load one beam in a Lucretia beam file as data for openPMD-beamphysics
    
    Parameters:
    ----------
    filename : str
               Lucretia '.mat' file name.
    ele_name : str
               name of the element at which the beam is located.
               An invalid name results in an error.
               If the beam file has one element, this only one beam is read. 
               Default: 'BEGINNING'
    t_ref : float, optional
            reference time of the beam in seconds. Default: 0.
    exclude_dead_particles : bool, optional
                             if True, excludes dead particles.  Default: True.

    
    Returns:
    ----------
    data: dict
        data for openPMD-beamphysics
    
    ----------

    
    Lucretia's format is described in:
    
        https://www.slac.stanford.edu/accel/ilc/codes/Lucretia/web/beam.html
        
    One Lucretia ".mat" file can include beams at multiple lattice elements.
    To find the beam at one element, one has to follow down this order of "fields":
    
        bstore >> ele_name >> Bunch >> x,
    
    in which x is a 6-to-Np array with:
    
        Lucretia x  = x in m
        Lucretia px = px/p in radian 
        Lucretia y  = y in m
        Lucretia py = py/p in radian 
        Lucretia z  = (t - t_ref)*c in m
        Lucretia p  = p in GeV/c
        
    Note that p is the total, not reference, momentum.
    
    To access valid element names in a Lucretia beam file, 
    use the helper function list_element_names(filename).
   
        dat = sio.loadmat('filename.mat')
        print(dat['bstore'].dtype)
        
    """

    ele_list = list_element_names(filename)
    if verbose:
        print(len(ele_list),'elements found in the file!')
    
    # Check if the element exists
    if (ele_name not in ele_list):
        raise ValueError('The provided element name '+ str(ele_name) +' does not exist in the file!')
    elif (len(ele_list) == 1):
        ele_name = ele_list[0]
        
    mdat = sio.loadmat(filename);
    
    coords = mdat['bstore'][ele_name][0,0]['Bunch'][0,0]['x'][0,0]
    charges = mdat['bstore'][ele_name][0,0]['Bunch'][0,0]['Q'][0,0][0]

    Np = coords.shape[1]

    x = coords[0]
    px_luc = coords[1] # normalized by total momentum 
    y = coords[2]
    py_luc = coords[3] # normalized by total momentum 
    z_luc = coords[4]
    ptot = coords[5]  # total momentum in GeV/c

    px = px_luc * ptot * 1e9 # in eV/c
    py = py_luc * ptot * 1e9
    pz = np.sqrt( (ptot * 1e9)**2 - px**2 - py**2 )

    t = z_luc / 299792458 + t_ref
       
    status = np.full(Np, 1)

    ix = np.where(ptot==0)
    status[ix] = 0
    n_dead = len(ix[0])

    if verbose:
        print(Np,'particles detected,', n_dead, 'found dead!')

    data = {
        'x':x,
        'px':px,
        'y':y,
        'py':py,
        'z':np.zeros(Np),
        'pz':pz,
        't':t, 
        'status':status, 
        'weight':charges, 
        'species':'electron'}
        
    if exclude_dead_particles and n_dead > 0:

        good = np.where(ptot>0)
        
        if verbose:
            print(f'Excluding {n_dead} dead particles')
            
        for k, v in data.items():
            if k == 'species':
                continue
            data[k] = data[k][good]
        
    return data


def write_lucretia(P, filePath, ele_name='BEGINNING', t_ref=0, stop_ix=None, verbose=True):
    """
    Write a ParticleGroup beam into a Lucretia beam file.
    
    Parameters:
    ----------
    P: ParticleGroup
               Particles to write. 
    
    filePath : str
               Lucretia '.mat' file name to be written to.
    ele_name : str
               name of the element at which the beam is located.
               Default: 'BEGINNING'
    t_ref : float, optional
            reference time of the beam in seconds. Default: 0.
    stop_ix : list of int, optional
              If provided, the length must equal to the number of particles.
              See Lucretia website for details.
              Default: None
    ----------
    
    Lucretia's format is described in:
    
        https://www.slac.stanford.edu/accel/ilc/codes/Lucretia/web/beam.html
      
    A general Lucretia beam file can include beams at multiple lattice elements.
    
    This routine only saves one beam at one element (name to be specified).
    Contents in the branches of the upper field structures are NOT defined.
    
    Lucretia beam follows:
    
        Lucretia x  = x in m
        Lucretia px = px/p in radian 
        Lucretia y  = y in m
        Lucretia py = py/p in radian 
        Lucretia z  = (t - t_ref)*c in m
        Lucretia p  = p in GeV/c
        
    Note that p is the total, not reference, momentum.
        
    """
    
    Np = P.n_particle

    ptot_nonzero = np.where( P.p==0, 1, P.p) # this prevents division by zero later
                                             # The replacement of "1" is dummy   
    x_luc = P.x
    px_luc = np.where( P.p==0, 0, P.px / ptot_nonzero )
    y_luc = P.y
    py_luc = np.where( P.p==0, 0, P.py / ptot_nonzero )
    z_luc = (P.t - t_ref) * 299792458
    ptot = P.p / 1E9 # total momentum in GeV/c

    if (np.all(stop_ix == None)):
        stop_ix = np.zeros(Np)
    else:
        if(len(stop_ix) != Np):
            print('Error: Length of stop_ix must equal to # particles !!')
        return
    
    # Form the lowest level of field structure
    l1 = np.array([x_luc,px_luc,y_luc,py_luc,z_luc,ptot])
    l2 = np.array([P.weight])
    l3 = np.array([stop_ix])
    
    # Wrapping upward in level fields 
    w1 = np.array([np.array([(l1, l2, l3)], dtype=[('x', 'O'), ('Q', 'O'), ('stop', 'O')])])
    w2 = np.array([np.array([(w1,)], dtype=[('Bunch', 'O')])])
    w3 = np.array([np.array([(w2,)], dtype=[(ele_name, 'O')])])
    w4 = {'bstore':w3}
    
    if verbose:
        print(f'writing {len(P)} particles in the Lucretia format to {filePath}')    
    
    sio.savemat(filePath, w4)
    
def list_element_names(filePath):
    """
    Find the element names in a Lucretia beam file
    
    Parameters:
    ----------
    filePath : str
               Lucretia '.mat' file name.

    
    Returns:
    ----------
    A list with the element names

    """
    dat = sio.loadmat(filePath)
    return list(dat['bstore'].dtype.fields)


from .units import dimension, dimension_name, SI_symbol, pg_units, c_light, e_charge
from .tools import decode_attrs, decode_attr


import h5py
import numpy as np

#-----------------------------------------
# General Utilities



#-----------------------------------------
# Records, components, units

particle_record_components = {
    'branchIndex':None,
    'chargeState':None,
    'electricField':['x', 'y', 'z'],
    'elementIndex':None,
    'magneticField':['x', 'y', 'z'],
    'locationInElement': None,
    'momentum':['x', 'y', 'z'],
    'momentumOffset':['x', 'y', 'z'],
    'photonPolarizationAmplitude':['x', 'y'],
    'photonPolarizationPhase':['x', 'y'],
    'sPosition':None,
    'totalMomentum':None,
    'totalMomentumOffset':None,
    #'particleCoordinatesToGlobalTransformation': ??
    'particleStatus':None,
    'pathLength':None,
    'position':['x', 'y', 'z'],
    'positionOffset':['x', 'y', 'z'],
    'spin':['x', 'y', 'z', 'theta', 'phi', 'psi'],
    'time':None,
    'timeOffset':None,
    'velocity':['x', 'y', 'z'],
    'weight':None
}

field_record_components = {
    'electricField':['x', 'y', 'z', 'r', 'theta'],
    'magneticField':['x', 'y', 'z', 'r', 'theta']
}


# Expected unit dimensions for particle and field records
expected_record_unit_dimension = {
    'branchIndex':dimension('1'),
    'chargeState':dimension('1'),
    'electricField':dimension('electric_field'),
    'magneticField':dimension('magnetic_field'),
    'elementIndex':dimension('1'),
    'locationInElement': dimension('1'),
    'momentum':dimension('momentum'),
    'momentumOffset':dimension('momentum'),
    'photonPolarizationAmplitude':dimension('electric_field'),
    'photonPolarizationPhase':dimension('1'),
    'sPosition':dimension('length'),
    'totalMomentum':dimension('momentum'),
    'totalMomentumOffset':dimension('momentum'),
    #'particleCoordinatesToGlobalTransformation': ??
    'particleStatus':dimension('1'),
    'pathLength':dimension('length'),
    'position':dimension('length'),
    'positionOffset':dimension('length'),
    'spin':dimension('1'),
    'time':dimension('time'),
    'timeOffset':dimension('time'),
    'velocity':dimension('velocity'),
    'weight':dimension('charge')
}

# Convenient aliases for components
component_from_alias = {
   # 'x':'position/x',
   # 'y':'position/y',
   # 'z':'position/z',
   # 'px':'momentum/x',
   # 'py':'momentum/y',
   # 'pz':'momentum/z',
    't':'time',
    'weight':'weight',
    'status':'particleStatus'
}
# Aliases for particles and fields
for g, prefix in zip(['position', 'momentum', 'electricField', 'magneticField'], 
                     ['', 'p', 'E', 'B']):
    for c in ['x', 'y', 'z', 'r', 'theta']:
        alias = prefix+c
        component_from_alias[alias] = g+'/'+c      
# Inverse
component_alias = {v:k for k,v in component_from_alias.items()}



def particle_paths(h5, key='particlesPath'):
    """
    Uses the basePath and particlesPath to find where openPMD particles should be
    
    """
    basePath = h5.attrs['basePath'].decode('utf-8')
    particlesPath = h5.attrs[key].decode('utf-8')
    
    if '%T' not in basePath:
        return [basePath+particlesPath]
    path1, path2 = basePath.split('%T')
    tlist = list(h5[path1])
    paths =  [path1+t+path2+particlesPath for t in tlist]
    return paths


def field_paths(h5, key='externalFieldPath'):
    """
    Looks for the External Fields
    
    """
    if key not in h5.attrs:
        return []
    
    fpath = h5.attrs[key].decode('utf-8')
    
    if '%T' not in fpath:
        return [fpath]
    
    path1 = fpath.split('%T')[0]
    tlist = list(h5[path1])
    paths =  [path1+t for t in tlist]
    return paths





def is_constant_component(h5):
    """
    Constant record component should have 'value' and 'shape'
    """
    return 'value' and 'shape' in h5.attrs

def constant_component_value(h5):
    """
    Constant record component should have 'value' and 'shape'
    """
    unitSI = h5.attrs['unitSI']
    val = h5.attrs['value']
    if  unitSI == 1.0:
        return val
    else:
        return val*unitSI

def component_unit_dimension(h5):
    """
    Return the unit dimension tuple
    """
    return tuple(h5.attrs['unitDimension'])
    
def component_data(h5, slice = slice(None), unit_factor=1):
    """
    Returns a numpy array from an h5 component.
    
    Determines wheter a component has constant data, or array data, and returns that. 
    
    An optional slice allows parts of the array to be retrieved. 
    
    This checks for a gridDataOrder attribute: F or C. If F, the np array is transposed. 
    
    Unit factor is an additional factor to convert from SI units to output units. 
    
    """

    # look for unitSI factor. 
    if 'unitSI' in h5.attrs:
        factor = h5.attrs['unitSI']
    else:
        factor = 1
    
    # Additional conversion factor
    if unit_factor:
        factor *= unit_factor
        
    if is_constant_component(h5):
        dat = np.full(h5.attrs['shape'], h5.attrs['value'])[slice]
    
    # Check multidimensional for data ordering
    elif len(h5.shape) > 1:        
        
        # Check for Fortran order
        if 'gridDataOrder' in h5.attrs and decode_attr(h5.attrs['gridDataOrder'])=='F':
            
            if isinstance(slice, tuple):
                # Need to transpose the slice ordering
                slice = slice[::-1]
    
            # Retrieve dataset and transpose for C order
            dat = h5[slice]
            dat = np.transpose(dat)
        else:
            # C-order
            dat = h5[slice]
            
    # 1-D array
    else:
        dat = h5[slice]

    if factor != 1:
        dat *= factor
        
    return dat

 
def offset_component_name(component_name):
    """
    Many components can also have an offset, as in:
    
        position/x
        positionOffset/c

    Return the appropriate name.
    """
    x = component_name.split('/')
    if len(x) == 1:
        return x[0]+'Offset'
    else:
        return x[0]+'Offset/'+x[1]


def particle_array(h5, component, slice=slice(None), include_offset=True):
    """
    Main routine to return particle arrays in fixed units.
    All units are SI except momentum, which will be in eV/c. 
    
    Example:
        particle_array(h5['data/00001/particles/'], 'px')
        Will return the momentum/x + momentumOffset/x in eV/c. 
        
        
    """

    # Handle aliases
    if component in component_from_alias:
        component = component_from_alias[component]

    if component in ['momentum/x', 'momentum/y', 'momentum/z']:
        unit_factor = (c_light / e_charge  ) # convert J/(m/s) to eV/c
    else:
        unit_factor = 1.0

    # Get data
    dat = component_data(h5[component], slice = slice, unit_factor=unit_factor)
        
        
    # Look for offset component
    ocomponent = offset_component_name(component)
    if include_offset and ocomponent in h5 :
        offset =  component_data(h5[ocomponent], slice = slice, unit_factor=unit_factor)
        dat += offset
        
        
    return dat
        
    

    
def all_components(h5):
    """
    Look for possible components in a particle group
    """
    components = []
    for record_name in h5:
        if record_name not in particle_record_components:
            continue
    
        # Look for components
        possible_components = particle_record_components[record_name]
        
        if not possible_components:
            # Record is a component
            components.append(record_name)
        else:
            g = h5[record_name]
            for cname in possible_components:
                if cname in g:
                    components.append(record_name+'/'+cname)
    
    return components



def component_str(particle_group, name):
    """
    Informational string from a component in a particle group (h5)
    """
    
    g = particle_group[name]
    record_name = name.split('/')[0]
    expected_dimension = expected_record_unit_dimension[record_name]
    this_dimension =  component_unit_dimension(g)
    dname = dimension_name(this_dimension)
    symbol = SI_symbol[dname]
  
    s = name+' '
    
    if is_constant_component(g):
        val = constant_component_value(g)
        shape = g.attrs['shape']
        s += f'[constant {val} with shape {shape}]'
    else:
        s += '['+str(len(g))+' items]'
        
    if symbol != '1':
        s += f' is a {dname} with units: {symbol}'
        
    if expected_dimension != this_dimension:   
        
        s +=', but expected units: '+ SI_symbol[dimension_name(expected_dimension)]
    
    return s






#----------------------------------
# Fields

required_field_attrs = [
    # strings
    'eleAnchorPt', 'gridGeometry', 'axisLabels',
    # reals and ints
    'gridLowerBound', 'gridOriginOffset', 'gridSpacing', 'gridSize', 'harmonic'
]            
    
# Dict with options    
optional_field_attrs = {
    'name':None,
    'gridCurvatureRadius':None,
    'fundamentalFrequency':0,
    'RFphase':0,
    'fieldScale':1.0,
    'masterParameter':None
}

    
def load_field_attrs(attr, verbose=False):
    """
    Loads FieldMesh required and optional attributes from a dict_like object.
    
    Non-standard attributes will be collected in an 'other' dict.
    
    Returns dicts:
        attrs, other
    
    """
    # Get all attrs. Will pop. 
    a = dict(attr)

    attrs = {}
    other = {}
    
    # Required 
    for k in required_field_attrs:
        attrs[k] = a.pop(k)

    # Optional, filling in some defaults
    for k in optional_field_attrs:
        if k in a:
            attrs[k] = a.pop(k)
        else:
            v = optional_field_attrs[k]
            if v is not None:
                attrs[k] = v
        
    # Collect other. 
    for k, v in a.items():
        other[k]= v
        if verbose: 
            print('Nonstandard attr:', k, v)
                
    # Decode            
    attrs  = decode_attrs(attrs)  
            
    # Error checking
    #if attrs['harmonic'] > 0:
    #    assert 'fundamentalFrequency' in attrs, 'fundamentalFrequency required if harmonic > 0'
    
    return attrs, other





from ..statistics import twiss_calc

from h5py import File
import numpy as np



# Genesis 1.3 
#-------------
# Version 2 routines


def genesis2_beam_data1(pg):
    """
    
    Calculate statistics of a single particlegroup, 
    for use in a Genesis 1.3 v2 beam file
    
    Returns a dict of:
        zpos or tpos      : z or t of the mean in m or s
        curpeak           : current in A
        gamma0            : average relativistic gamma (dimensionless)
        emitx, emity      : Normalized emittances in m*rad
        rxbeam, rybeam    : sigma_x, sigma_y in m
        xbeam, ybeam      : <x>, <y> in m
        pxbeam, pybeam    : beta_x*gamma, beta_y*gamma (dimensionless)
        alpha_x, alpha_y  : Twiss alpha  (dinensionless)
    """
    
    d = {}
    
    # Handle z or t
    if len(set(pg.z)) == 1:
        # should be different t
        d['tpos'] = pg['mean_t']
    elif len(set(pg.t)) == 1:
         #should be different z
        d['zpos'] = pg['mean_z']
    else:
        raise ValueError(f'{pg} has mixed t and z coordinates.')

    d['curpeak'] = pg['average_current']     
    d['gamma0'] = pg['mean_gamma']    
    d['delgam'] = pg['sigma_gamma']
    d['emitx'] = pg['norm_emit_x']
    d['emity'] = pg['norm_emit_y']        
    d['rxbeam'] = pg['sigma_x']
    d['rybeam'] = pg['sigma_y']
    d['xbeam'] = pg['mean_x']
    d['ybeam'] = pg['mean_y']
    
    d['pxbeam'] = pg['mean_px']/pg.mass # beta_x*gamma
    d['pybeam'] = pg['mean_py']/pg.mass # beta_y*gamma
    
    # Twiss, for alpha
    twiss_x = twiss_calc(pg.cov('x', 'xp'))
    twiss_y = twiss_calc(pg.cov('y', 'yp'))
    d['alphax'] = twiss_x['alpha'] 
    d['alphay'] = twiss_y['alpha'] 
        
    return d


def genesis2_beam_data(pg, n_slice=None):
    """
    
    Slices a particlegroup into n_slice and forms the beam columns. 
    
    n_slice is the number of slices. If not given, the beam will be divided 
    so there are 100 particles in each slice. 
    
    Returns a dict of beam_columns, for use with write_genesis2_beam_file
    
    See: genesis2_beam_data1
    """
    
    
    # Handle z or t
    if len(set(pg.z)) == 1:
        # should be different t
        slice_key = 't'
    elif len(set(pg.t)) == 1:
         #should be different z
        slice_key = 'z'
    else:
        raise ValueError(f'{pg} has mixed t and z coordinates.')    
    
    # Automatic slicing
    if not n_slice:
        n_slice = len(pg)//100
    
    # Slice (split)
    pglist = pg.split(n_slice, key=slice_key)
    
    d = {}
    # Loop over the slices
    for pg in pglist:
        data1 = genesis2_beam_data1(pg)
        for k in data1:
            if k not in d:
                d[k] = []
            d[k].append(data1[k])
    for k in d:
        d[k] = np.array(d[k])
    
    return d
    

def write_genesis2_beam_file(fname, beam_columns, verbose=False):
    """
    Writes a Genesis 1.3 v2 beam file, using a dict beam_columns
    
    The header will be written as:
    ? VERSION=1.0
    ? SIZE=<length of the columns>
    ? COLUMNS <list of columns
    <data>

    This is a copy of the lume-genesis routine:
    
    genesis.writers.write_beam_file
    
    """
    
    # Get size
    names = list(beam_columns)
    size = len(beam_columns[names[0]])
    header=f"""? VERSION=1.0
? SIZE={size}
? COLUMNS {' '.join([n.upper() for n in names])}"""
    
    dat = np.array([beam_columns[name] for name in names]).T

    np.savetxt(fname, dat, header=header, comments='', fmt='%1.8e') # Genesis can't read format %1.16e - lines are too long?
    
    if verbose:
        print('Beam written:', fname)
    
    return header



#-------------
# Version 4 routines

def write_genesis4_distribution(particle_group,
                             h5file,
                             verbose=False):
    """
    
    
    Cooresponds to the `import distribution` section in the Genesis4 manual. 
    
    Writes datesets to an h5 file:
    
    h5file: str or open h5 handle
    
    Datasets
        x is the horizontal coordinate in meters
        y is the vertical coordinate in meters
        xp = px/pz is the dimensionless trace space horizontal momentum
        yp = py/pz is the dimensionless trace space vertical momentum
        t is the time in seconds
        p  = relativistic gamma*beta is the total momentum divided by mc

    
        These should be the same as in .interfaces.elegant.write_elegant
        
        
    If particles are at different z, they will be drifted to the same z, 
    because the output should have different times. 
    
    """


    if isinstance(h5file, str):
        h5 = File(h5file, 'w')
    else:
        h5 = h5file
        
    if len(set(particle_group.z)) > 1:
        if verbose:
            print('Drifting particles to the same z')
        # Work on a copy, because we will drift
        P = particle_group.copy()
        # Drift to z. 
        P.drift_to_z()        
        
    
    else:
        P = particle_group

        
    for k in ['x', 'xp', 'y', 'yp', 't']:
        h5[k] = P[k]
        
    # p is really beta*gamma    
    h5['p'] = P['p']/P.mass
    
    
    if verbose:
        print(f'Datasets x, xp, y, yp, t, p written to: {h5file}')
        

   



from pmd_beamphysics.units import dimension, dimension_name, SI_symbol, pg_units

from pmd_beamphysics.readers import component_data, expected_record_unit_dimension, field_record_components, decode_attrs, field_paths, component_from_alias, load_field_attrs, component_alias

from pmd_beamphysics.writers import write_pmd_field, pmd_field_init

from pmd_beamphysics import tools

from pmd_beamphysics.plot import plot_fieldmesh_cylindrical_2d

from pmd_beamphysics.interfaces.superfish import write_superfish_t7, read_superfish_t7
from pmd_beamphysics.interfaces.gpt import write_gpt_fieldmesh

from h5py import File
import numpy as np
from copy import deepcopy
import os


c_light = 299792458.


#-----------------------------------------
# Classes


axis_labels_from_geometry = {
    'cartesian': ('x', 'y', 'z'),
    'cylindrical': ('r', 'theta', 'z')
}


class FieldMesh:
    """
    Class for openPMD External Field Mesh data.
    
    Initialized on on openPMD beamphysics particle group:
        h5 = open h5 handle, or str that is a file
        data = raw data
        
    The required data is stored in ._data, and consists of dicts:
        'attrs'
        'components'
    
    Component data is always 3D.
    
    Initialization:
        From openPMD-beamphysics HDF5 file:
            FieldMesh('file.h5')
        From data dict:
            FieldMesh(data=data)
    
    Derived properties:
                
        .r, .theta, .z
        .Br, .Btheta, .Bz
        .Er, .Etheta, .Ez
        .E, .B
        
        .phase
        .scale
        .factor
        
        .harmonic
        .frequency
        
        .shape
        .geometry
        .mins, .maxs, .deltas
        .meshgrid
        .dr, .dtheta, .dz
    
    Booleans:
        .is_pure_electric
        .is_pure_magnetic
        .is_static
    
    Units and labels
        .units
        .axis_labels
    
    Plotting:
        .plot
    
    Writers
        .write
        .write_gpt
        .write_superfish
        
    Readers (class methods):
        .from_superfish
    

    

    
    """
    def __init__(self, h5=None, data=None):
    
        if h5:
            # Allow filename
            if isinstance(h5, str):
                fname = os.path.expandvars(os.path.expanduser(h5))
                assert os.path.exists(fname), f'File does not exist: {fname}'
  
                with File(fname, 'r') as hh5:
                    fp = field_paths(hh5)
                    assert len(fp) == 1, f'Number of field paths in {h5}: {len(fp)}'
                    data = load_field_data_h5(hh5[fp[0]])

            else:
                pass
        else:
            print('loading data')
            data = load_field_data_dict(data)
            
        # Internal data
        self._data = data
            
        # Aliases (Do not set these! Set via slicing: .Bz[:] = 0
        #for k in self.components:
        #    alias = component_alias[k]
        #    self.__dict__[alias] =  self.components[k]
            
           
    # Direct access to internal data        
    @property
    def attrs(self):
        return self._data['attrs']
    
    @property
    def components(self):
        return self._data['components']  
    
    @property
    def data(self):
        return self._data
    
    
    # Conveniences 
    @property
    def shape(self):
        return tuple(self.attrs['gridSize'])
    
    @property
    def geometry(self):
        return self.attrs['gridGeometry']
    
    @property
    def scale(self):
        return self.attrs['fieldScale']    
    @scale.setter
    def scale(self, val):
        self.attrs['fieldScale']  = val
        
    @property
    def phase(self):
        """
        Returns the complex argument phi = -2*pi*RFphase
        to multiply the oscillating field by. 
        
        Can be set. 
        """
        return -self.attrs['RFphase']*2*np.pi
    @phase.setter
    def phase(self, val):
        """
        Complex argument in radians
        """
        self.attrs['RFphase']  = -val/(2*np.pi)    
   
    @property
    def factor(self):
        """
        factor to multiply fields by, possibly complex.
        
        factor = scale * exp(i*phase)
        """
        return np.real_if_close(self.scale * np.exp(1j*self.phase))           
    
    @property
    def axis_labels(self):
        """
        
        """
        return axis_labels_from_geometry[self.geometry]
    
    def axis_index(self, key):
        """
        Returns axis index for a named axis label key.
        
        Example:
            .axis_labels == ('x', 'y', 'z')
            .axis_index('z') returns 2
        """
        for i, name in enumerate(self.axis_labels):
            if name == key:
                return i
        raise ValueError(f'Axis not found: {key}')
    
    @property
    def coord_vecs(self):
        """
        Uses gridSpacing, gridSize, and gridOriginOffset to return coordinate vectors.
        """
        return [np.linspace(x0, x1, nx) for x0, x1, nx in zip(self.mins, self.maxs, self.shape)]
        
    def coord_vec(self, key):
        """
        Gets the coordinate vector from a named axis key. 
        """
        i = self.axis_index(key)
        return np.linspace(self.mins[i], self.maxs[i], self.shape[i])
        
    @property 
    def meshgrid(self):
        """
        Usses coordinate vectors to produce a standard numpy meshgrids. 
        """
        vecs = self.coord_vecs
        return np.meshgrid(*vecs, indexing='ij')
        
    
    
    @property 
    def mins(self):
        return np.array(self.attrs['gridOriginOffset'])
    @property
    def deltas(self):
        return np.array(self.attrs['gridSpacing'])
    @property
    def maxs(self):
        return self.deltas*(np.array(self.attrs['gridSize'])-1) + self.mins      
    
    @property
    def frequency(self):
        if self.is_static:
            return 0
        else:
            return self.attrs['harmonic']*self.attrs['fundamentalFrequency']
    
    
    # Logicals
    @property
    def is_pure_electric(self):
        """
        Returns True if there are no non-zero mageneticField components
        """
        klist = [key for key in self.components if not self.component_is_zero(key)]
        return all([key.startswith('electric') for key in klist])
    # Logicals
    @property
    def is_pure_magnetic(self):
        """
        Returns True if there are no non-zero electricField components
        """
        klist = [key for key in self.components if not self.component_is_zero(key)]
        return all([key.startswith('magnetic') for key in klist])
            

        
    @property
    def is_static(self):
        return  self.attrs['harmonic'] == 0
    

    
    def component_is_zero(self, key):
        """
        Returns True if all elements in a component are zero.
        """
        a = self[key]
        return not np.any(a)
                

    # Plotting
    # TODO: more general plotting
    def plot(self, component=None, time=None, axes=None, cmap=None, return_figure=False, **kwargs):
        
        
        return plot_fieldmesh_cylindrical_2d(self,
                                             component=component,
                                             time=time,
                                             axes=axes,
                                             return_figure=return_figure,
                                             cmap=cmap, **kwargs)
    
    
    def units(self, key):
        """Returns the units of any key"""
        
        # Strip any operators
        _, key = get_operator(key)
        
        # Fill out aliases 
        if key in component_from_alias:
            key = component_from_alias[key]
        elif key == 'E':
            key = 'electricField'
        elif key == 'B':
            key = 'magneticField'            
        
        return pg_units(key)    
    
    # openPMD    
    def write(self, h5, name=None):
        """
        Writes openPMD-beamphysics format to an open h5 handle, or new file if h5 is a str.
        
        """
        if isinstance(h5, str):
            fname = os.path.expandvars(os.path.expanduser(h5))
            h5 = File(fname, 'w')
            pmd_field_init(h5, externalFieldPath='/ExternalFieldPath/%T/')
            g = h5.create_group('/ExternalFieldPath/1/')
        else:
            g = h5
    
        write_pmd_field(g, self.data, name=name)   
        
       
    def write_gpt(self, filePath, asci2gdf_bin=None, verbose=True):
        """
        Writes a GPT field file. 
        """
    
        return write_gpt_fieldmesh(self, filePath, asci2gdf_bin=asci2gdf_bin, verbose=verbose)
    
    # Superfish
    def write_superfish(self, filePath, verbose=False):
        """
        Write a Superfish T7 file. 
        
        For static fields, a Poisson T7 file is written.
        
        For dynamic (harmonic /= 0) fields, a Fish T7 file is written
        """
        return write_superfish_t7(self, filePath, verbose=verbose)
            
            
            
    @classmethod
    def from_superfish(cls, filename, type='electric', geometry='cylindrical'):
        """
        Class method to parse a superfish T7 style file.
        """        
        data = read_superfish_t7(filename, type=type, geometry=geometry)
        c = cls(data=data)
        return c               
        
        
        
    def __eq__(self, other):
        """
        Checks that all attributes and component internal data are the same
        """
        
        if not tools.data_are_equal(self.attrs, other.attrs):
            print('here')
            return False
        
        return tools.data_are_equal(self.components, other.components)
  

 #  def __setattr__(self, key, value):
 #      print('a', key)
 #      if key in component_from_alias:
 #          print('here', key)
 #          comp = component_from_alias[key]
 #          if comp in self.components:
 #              self.components[comp] = value

 #  def __getattr__(self, key):
 #      print('a')
 #      if key in component_from_alias:
 #          print('here', key)
 #          comp = component_from_alias[key]
 #          if comp in self.components:
 #              return self.components[comp]
    
    
    

    
    def scaled_component(self, key):
        """
        
        Retruns a component scaled by the complex factor
            factor = scale*exp(i*phase)
            
            
        """

        if key in self.components:
            dat = self.components[key] 
        # Aliases
        elif key in component_from_alias:
            comp = component_from_alias[key]
            if comp in self.components:
                dat = self.components[comp]   
            else:
                # Component not present, make zeros
                return np.zeros(self.shape)
        else:
            raise ValueError(f'Component not available: {key}')
        
        # Multiply by scale factor
        factor = self.factor      
        
        if factor != 1:
            return factor*dat
        else:
            return dat

    # Convenient properties
    # TODO: Automate this?
    @property
    def r(self):
        return self.coord_vec('r')
    @property
    def theta(self):
        return self.coord_vec('theta')
    @property
    def z(self):
        return self.coord_vec('z')    
    
    @property
    def dr(self):
        return self.deltas[self.axis_index('r')]
    @property
    def dtheta(self):
        return self.deltas[self.axis_index('theta')]
    @property
    def dz(self):
        return self.deltas[self.axis_index('z')]    
    
    # TODO: Cartesian components, checking geometry
    @property
    def Br(self):
        return self.scaled_component('Br')
    @property
    def Btheta(self):
        return self.scaled_component('Btheta')
    @property
    def Bz(self):
        return self.scaled_component('Bz')
    @property
    def Er(self):
        return self.scaled_component('Er')
    @property
    def Etheta(self):
        return self.scaled_component('Etheta')
    @property
    def Ez(self):
        return self.scaled_component('Ez')    
    
    @property
    def B(self):
        if self.geometry=='cylindrical':
            if self.is_static:
                return np.hypot(self['Br'], self['Bz'])
            else:
                return np.abs(self['Btheta'])
        else:
            raise ValueError(f'Unknown geometry: {self.geometry}')    
          
    @property
    def E(self):
        if self.geometry=='cylindrical':
            return np.hypot(np.abs(self['Er']), np.abs(self['Ez']))
        else:
            raise ValueError(f'Unknown geometry: {self.geometry}')    
        
    def __getitem__(self, key):
        """
        Returns component data from a key
        
        If the key starts with:
            re_
            im_
            abs_
        the appropriate numpy operator is applied.
        
        
        
        """
        
        # 
        if key in ['r', 'theta', 'z']:
            return self.coord_vec(key)
            
            
        # Raw components
        if key in self.components:
            return self.components[key]
       
        # Check for operators
        operator, key = get_operator(key)
        
        # Scaled components
        if key == 'E':
            dat = self.E
        elif key == 'B':
            dat =  self.B
        else:
            dat = self.scaled_component(key)        
                   
        if operator:
            dat = operator(dat)
            
        return dat
        

        
        
    def __repr__(self):
        memloc = hex(id(self))
        return f'<FieldMesh with {self.geometry} geometry and {self.shape} shape at {memloc}>'        

    

def get_operator(key):
    """
    Check if a key starts with re_, im_, abs_
    
    returns operator, newkey
    """
    # Check for simple operators
    if key.startswith('re_'):
        operator = np.real
        newkey = key[3:]
    elif key.startswith('im_'):
        operator = np.imag
        newkey = key[3:]
    elif key.startswith('abs_'):
        operator = np.abs
        newkey = key[4:]            
    else:
        operator = None 
        newkey = key
    
    return operator, newkey
    
            
def load_field_data_h5(h5, verbose=True):
    """
    
    
    If attrs['dataOrder'] == 'F', will transpose.
    
    If attrs['harmonic'] == 0, components will be cast to real by np.real
    
    Returns:
        data dict
    """
    data = {'components':{}}

    # Load attributes
    attrs, other = load_field_attrs(h5.attrs, verbose=verbose)
    attrs.update(other)
    data['attrs'] = attrs
    
    # Loop over records and components
    for g, comps in field_record_components.items():
        if g not in h5:
            continue
        
        # Get the full openPMD unitDimension 
        required_dim = expected_record_unit_dimension[g]
                
        for comp in comps:
            if comp not in h5[g]:
                continue
            name = g+'/'+comp
            cdat = component_data(h5[name])
                
            # Check dimensions
            dim = h5[name].attrs['unitDimension']
            assert np.all(dim == required_dim), f'{name} with dimension {required_dim} expected for {name}, found: {dim}'
            
            # Check shape
            s1 = tuple(attrs['gridSize'])
            s2 = cdat.shape
            assert s1 == s2, f'Expected shape: {s1} != found shape: {s2}'
            
            # Static fields should be real
            if attrs['harmonic'] == 0:
                cdat = np.real(cdat)
            
            # Finally set
            
            data['components'][name] = cdat        
            
    
    return data

def load_field_data_dict(data_dict, verbose=True):
    """
    Similar to load_field_data_h5, but from a dict. 
    
    This cannot do unit checking. 
    """
    
    # The output dict
    data = {}
    
    # Load attributes
    attrs, other = load_field_attrs(data_dict['attrs'], verbose=verbose)
    attrs.update(other)
    data['attrs'] = attrs    
    
    # Go through components. Allow aliases
    comp = data['components'] = {}
    for k, v in data_dict['components'].items():
        if k in component_alias:
            comp[k] = v
        elif k in component_from_alias:
            k = component_from_alias[k]
            assert k not in data
            comp[k] = v
        else:
            raise ValueError(f'Unallowed component: {k}')
    
    
    return data






import numpy as np


def write_astra(particle_group,
                outfile,
                verbose=False,
                species='electron',
                probe=False):
    """
    Writes Astra style particles from particle_group type data.
    
    For now, the species must be electrons. 
    
    If probe, the six standard probe particles will be written. 
    """

    def vprint(*a, **k):
        if verbose:
            print(*a, **k)
    
    vprint(f'writing {particle_group.n_particle} particles to {outfile}')

    assert species == 'electron' # TODO: add more species
    
    # number of lines in file
    size = particle_group.n_particle + 1 # Allow one for reference particle
    i_start = 1 # Start for data particles
    if probe:
        # Add six probe particles, according to the manual
        size += 6
        i_start += 6
    

    # Astra units and types
    #units = ['m', 'm', 'm', 'eV/c', 'eV/c', 'eV/c', 'ns', 'nC']
    names = ['x', 'y', 'z', 'px', 'py', 'pz', 't', 'q', 'index', 'status']
    types = 8*[np.float] + 2*[np.int8]

    
    # Reference particle
    ref_particle = {'q':0}
    sigma = {}
    for k in ['x', 'y', 'z', 'px', 'py', 'pz', 't']:
        ref_particle[k] = particle_group.avg(k)
        sigma[k] =  particle_group.std(k)
    ref_particle['t'] *= 1e9 # s -> nS
        
    # Make structured array
    dtype = np.dtype(list(zip(names, types)))
    data = np.zeros(size, dtype=dtype)
    for k in ['x', 'y', 'z', 'px', 'py', 'pz', 't']:
        data[k][i_start:] = getattr(particle_group, k)
    data['t'] *= 1e9 # s -> nS
    data['q'][i_start:] = particle_group.weight*1e9 # C -> nC
    
    # Set these to be the same
    data['index'] = 1    # electron
    data['status'] = -1  # Particle at cathode
    
    # Subtract off reference z, pz, t
    for k in ['z', 'pz', 't']:
        data[k] -= ref_particle[k]
        
    # Put ref particle in first position
    for k in ref_particle:
        data[k][0] = ref_particle[k]
    
    # Optional: probes, according to the manual
    if probe:
        data[1]['x'] = 0.5*sigma['x'];data[1]['t'] =  0.5*sigma['t']
        data[2]['y'] = 0.5*sigma['y'];data[2]['t'] = -0.5*sigma['t']
        data[3]['x'] = 1.0*sigma['x'];data[3]['t'] =  sigma['t']
        data[4]['y'] = 1.0*sigma['y'];data[4]['t'] = -sigma['t']
        data[5]['x'] = 1.5*sigma['x'];data[5]['t'] =  1.5*sigma['t']
        data[6]['y'] = 1.5*sigma['y'];data[6]['t'] = -1.5*sigma['t']        
        data[1:7]['status'] = -3
        data[1:7]['pz'] = 0 #? This is what the Astra Generator does
    
    # Save in the 'high_res = T' format
    np.savetxt(outfile, data, fmt = ' '.join(8*['%20.12e']+2*['%4i']))




from .particles import ParticleGroup, single_particle
from .fields.fieldmesh import FieldMesh
from .readers import particle_paths
from .writers import pmd_init

from . import _version
__version__ = _version.get_versions()['version']



import numpy as np
   
def write_opal(particle_group,           
               outfile,
               dist_type = 'emitted',
               verbose=False): 

    """
    OPAL's ASCII format is described in:
    
        https://gitlab.psi.ch/OPAL/Manual-2.2/wikis/distribution
    
    outfile is the name out the ASCII file to be written.
    
    dist_type is one of:
    
        'emitted' : The longitudinal dimension is 
                    described by time, and the particles are emitted 
                    from the cathode plane over a set number of 'emission steps'.
                    
        'injected': the longiudinal dimension is 
                    described by z, and the particles are born instaneously 
                    in the simulation at time step 0.

    """

    n = particle_group.n_particle            
    x = particle_group.x
    y = particle_group.y
    
    # Get longitudinal coordinate
    if dist_type == 'emitted':
        
        # Check that z are all the same
        unique_z = np.unique(particle_group['z'])
        assert len(unique_z) == 1, 'All particles must be a the same z position'
        
        z = particle_group.t
        
    elif dist_type == 'injected':
        
        # Check that t are all the same
        unique_t = np.unique(particle_group['t'])
        assert len(unique_t) == 1, 'All particles must be a the same time'
        
        z = particle_group.z

    else:
        raise ValueError(f'unknown dist_type: {dist_type}')
    
    gamma = particle_group.gamma
    GBx = gamma*particle_group.beta_x
    GBy = gamma*particle_group.beta_y
    GBz = gamma*particle_group.beta_z
    
    header=str(n)
    dat = np.array([x, GBx, y, GBy, z, GBz]).T
    
    if verbose:
        print(f'writing {dist_type} {n} particles to {outfile}')
    np.savetxt(outfile, dat, header=header, comments='', fmt = '%20.12e')
    
    
    
    
def opal_to_data(h5):
    """
    Converts an OPAL step to the standard data format for openPMD-beamphysics
    
    In OPAL, the momenta px, py, pz are gamma*beta_x, gamma*beta_y, gamma*beta_z. 
    
    TODO: More species. 
        
    """
    D = dict(h5.attrs)
    mc2 = D['MASS'][0]*1e9 # GeV -> eV 
    charge = D['CHARGE'][0] # total charge in C
    t = D['TIME'][0] # s
    ptypes = h5['ptype'][:]  # 0 = electron?
    
    # Not used: pref = D['RefPartP']*mc2  # 
    rref = D['RefPartR']
    
    n = len(ptypes)
    assert all(h5['ptype'][:] == 0)
    species = 'electron'
    status=1
    data = {
        'x':h5['x'][:] + rref[0],
        'y':h5['y'][:] + rref[1],
        'z':h5['z'][:] + rref[2],
        'px':h5['px'][:]*mc2,
        'py':h5['py'][:]*mc2,
        'pz':h5['pz'][:]*mc2,
        't': np.full(n, t),
        'status': np.full(n, status),
        'species':species,
        'weight': np.full(n, abs(charge)/n)
    }
    return data


        



from pmd_beamphysics.units import mec2

import numpy as np
import subprocess
import os

from h5py import File

def write_elegant(particle_group,           
               outfile,
               verbose=False): 

    """
    Elegant uses SDDS files. 
    
    Because elegant is an s-based code, particles are drifted to the center. 
    
    This routine writes an SDDS1 ASCII file, with a parameter
        Charge
   and columns
        't', 'x', 'xp', 'y', 'yp', 'p'        
    where 'p' is gamma*beta, in units: 
        elegant units are:
        s, m, 1, m, 1, 'mass*c'
    
    All weights must be the same. 

    """


    # Work on a copy, because we will drift
    P = particle_group.copy()

    
    # Drift to z. 
    P.drift_to_z()
    
    # Form data
    keys = ['t', 'x', 'xp', 'y', 'yp', 'p']
    dat = {}
    for k in keys:
        dat[k] = P[k]
    # Correct p, this is really gamma*beta    
    dat['p'] /= P.mass
         
    if verbose:
        print(f'writing {len(P)} particles to {outfile}')
        
    # Correct units for p, depending on the species
    species = particle_group.species
    if species in ['electron', 'positron']:
        p_units = 'units="m$be$nc", '
    elif species in ['proton']:
        p_units = 'units="m$bp$nc", '        
    else:
        p_units = ''

    # Note that the order of the columns matters below. 
    header = f"""SDDS1
! 
! Created using the openPMD-beamphysics Python package
! https://github.com/ChristopherMayes/openPMD-beamphysics
! species: {P['species']}
!
&parameter name=Charge, type=double, units=C, description="total charge in Coulombs" &end
&column name=t,  type=double, units=s, description="time in seconds" &end
&column name=x,  type=double, units=m, description="x in meters" &end
&column name=xp, type=double, description="px/pz" &end
&column name=y,  type=double, units=m, description="y in meters" &end
&column name=yp, type=double, description="py/pz" &end
&column name=p,  type=double, {p_units}description="relativistic gamma*beta" &end
&data mode=ascii &end
{P['charge']}
{len(P)}"""
    
    # Write ASCII
    outdat = np.array([dat[k] for k in keys]).T        
    np.savetxt(outfile, outdat, header=header, comments='', fmt = '%20.12e')    
      
    
    return outfile



def elegant_h5_to_data(h5, group='page1', species='electron'):
    """
    Converts elegant data from an h5 handle or file to data for openPMD-beamphysics
    
    Elegant h5 has datasets:
    
    x, xp, y, xp, p=beta*gamma, t
    m,  1,   m, 1, 1, s
    
    Momentum are reconstructed as:
    
    pz = p / sqrt(1+xp^2 + yp^2)
    px = xp * pz
    py = yp * pz
    
    In s-based codes, z=0 by definition. 
    
    For now, only electrons are allowed.
    
    Units are checked. 
    
    All particles are assumed to be live (status = 1)

    TODO: More species. 
        
    """
    
    # Allow for opening a file
    if isinstance(h5, str):
        assert os.path.exists(h5), f'File does not exist: {h5}'
        h5 = File(h5, 'r')
        
    if group:
        g = h5[group]
    else:
        g = h5
        
    assert species=='electron', f'{species} not allowed yet. Only electron is implemented.'    
    mc2 = mec2
        
    # These should exist
    col = g['columns']
    par = g['parameters']
    
    p = col['p'][:]*mc2
    xp = col['xp'][:]
    yp = col['yp'][:]
    pz = p/np.sqrt(1 + xp**2 + yp**2)
    px = xp * pz
    py = yp * pz
    

    # Check charge unit
    charge = par['Charge'][0]
    charge_unit = par['Charge'].attrs['units']
    assert charge_unit == b'C', 'Expected C as unit for Charge'
    
    # Check dataset units
    expected_units = {
        'p':b'm$be$nc',
        'xp':b'',
        'yp':b'',
        'x':b'm',
        'y':b'm',
        't':b's'
    }    
    for c, v in expected_units.items():
        u = col[c].attrs['units']
        assert u==v, f'Dataset {c} units expected to have {v}, but have {u}'
    
    # number of particles
    n = len(p)
    
    status=1
    data = {
        'x':col['x'][:],
        'y':col['y'][:],
        'z':np.full(n, 0),
        'px':px,
        'py':py,
        'pz':pz,
        't': col['t'][:],
        'status': np.full(n, status),
        'species':species,
        'weight': np.full(n, abs(charge)/n),
        'id':  col['particleID'][:]
    }
    return data





def load_sdds(sddsfile, columns, sdds2plaindata_bin='sdds2plaindata', verbose=False):
    """
    Get tabular data from SDDS file
    
    """
    
    outfile = sddsfile+'_table'
    cmd0 = [sdds2plaindata_bin, sddsfile, outfile, '-noRowCount','-outputMode=ascii']
    
    cmd = cmd0 + [f'-col={c}' for c in columns] + ['-separator= ']
    
    if verbose:
        print('load_sdds command: ', ' '.join(cmd))
    
    output, error  = subprocess.Popen(
                    cmd, universal_newlines=True,
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
    if error:
        raise ValueError('load_sdds error: '+error)
    # Read table  
    rdat = np.loadtxt(outfile, ndmin=2)

    dat = {}
    for i, key in  enumerate(columns):
        dat[key] = rdat[:,i]
    
    # Cleanup
    os.remove(outfile)
    
    return dat


def elegant_to_data(sddsfile, charge=1.0, sdds2plaindata_bin='sdds2plaindata', species='electron', verbose=False):
    """
    Converts elegant SDDS data to data for openPMD-beamphysics.
    
    Similar to elegant_h5_to_data, which is better because it checks units and does not need the conversion to ASCII
    
    Requires the SDDS utility sdds2plaindata
    
    Columns in the file should be:
    
    x, xp, y, xp, p=beta*gamma, t
    m,  1,   m, 1, 1, s
    
    Momentum are reconstructed as:
    
    pz = p / sqrt(1+xp^2 + yp^2)
    px = xp * pz
    py = yp * pz
    
    In s-based codes, z=0 by definition. 
    
    For now, only electrons are allowed.
    
    
    All particles are assumed to be live (status = 1)

    TODO: More species. 
    
    Also see: elegant_h5_to_data
        
    """
    
    col = load_sdds(sddsfile, ['x', 'xp', 'y', 'yp', 't', 'p'], sdds2plaindata_bin=sdds2plaindata_bin, verbose=verbose)
    
        
    assert species=='electron', f'{species} not allowed yet. Only electron is implemented.'    
    mc2 = mec2
            
    p =  col['p']*mc2
    xp = col['xp']
    yp = col['yp']
    pz = p/np.sqrt(1 + xp**2 + yp**2)
    px = xp * pz
    py = yp * pz   
    
    # number of particles
    n = len(p)
    
    status=1
    data = {
        'x':col['x'],
        'y':col['y'],
        'z':np.full(n, 0),
        'px':px,
        'py':py,
        'pz':pz,
        't': col['t'],
        'status': np.full(n, status),
        'species':species,
        'weight': np.full(n, abs(charge)/n)
    }
    return data



"""


"""

from  pmd_beamphysics.units import nice_array, nice_scale_prefix
from .statistics import slice_statistics
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.gridspec import GridSpec
import numpy as np
from copy import copy
CMAP0 = copy(plt.get_cmap('viridis'))
CMAP0.set_under('white')

CMAP1 = copy(plt.get_cmap('plasma'))

# For field legends
from mpl_toolkits.axes_grid1 import make_axes_locatable


def plt_histogram(a, weights=None, bins=40):
    """
    This produces the same plot as plt.hist
    
    For reference only
    
    Note that bins='auto', etc cannot be used if there are weights.
    """
    cnts, bins = np.histogram(a, weights=weights, bins=bins)
    plt.bar(bins[:-1] + np.diff(bins) / 2, cnts, np.diff(bins))

    

    
    
def slice_plot(particle_group, stat_key='sigma_x', n_slice=40, slice_key='z'):
    """
    Complete slice plotting routine. Will plot the density of the slice key on the right axis. 
    """
    
    x_key = 'mean_'+slice_key
    y_key = stat_key
    slice_dat = slice_statistics(particle_group, n_slice=n_slice, slice_key=slice_key,
                            keys=[x_key, y_key, 'ptp_'+slice_key, 'charge'])
    
    
    slice_dat['density'] = slice_dat['charge']/ slice_dat['ptp_'+slice_key]
    y2_key = 'density'
    fig, ax = plt.subplots()
    
    # Get nice arrays
    x, _, prex = nice_array(slice_dat[x_key])
    y, _, prey = nice_array(slice_dat[y_key])
    y2, _, prey2 = nice_array(slice_dat[y2_key])
    
    # Add prefix to units
    x_units = prex+particle_group.units(x_key).unitSymbol
    y_units = prey+particle_group.units(y_key).unitSymbol
    
    # Convert to Amps if possible
    y2_units = f'C/{particle_group.units(x_key)}'
    if y2_units == 'C/s':
        y2_units = 'A'
    y2_units = prey2+y2_units 
    
    # Labels
    ax.set_xlabel(f'{x_key} ({x_units})' )
    ax.set_ylabel(f'{y_key} ({y_units})' )
    
    # Main plot
    ax.plot(x, y, color = 'black')
    
    #ax.set_ylim(0, 1.1*ymax )

    ax2 = ax.twinx()
    ax2.set_ylabel(f'{y2_key} ({y2_units})' )
    ax2.fill_between(x, 0, y2, color='black', alpha = 0.2)  

    
    
def density_plot(particle_group, key='x', bins=None, **kwargs):
    """
    1D density plot. Also see: marginal_plot
    
    Example:
    
        density_plot(P, 'x', bins=100)   
    
    """
    
    if not bins:
        n = len(particle_group)
        bins = int(n/100)

    # Scale to nice units and get the factor, unit prefix
    x, f1, p1 = nice_array(particle_group[key])
    w = particle_group['weight']
    u1 = particle_group.units(key).unitSymbol
    ux = p1+u1
    
    labelx = f'{key} ({ux})'
    
    fig, ax = plt.subplots(**kwargs)
    
    hist, bin_edges = np.histogram(x, bins=bins, weights=w)
    hist_x = bin_edges[:-1] + np.diff(bin_edges) / 2
    hist_width =  np.diff(bin_edges)
    hist_y, hist_f, hist_prefix = nice_array(hist/hist_width)
    ax.bar(hist_x, hist_y, hist_width, color='grey')
    # Special label for C/s = A
    if u1 == 's':
        _, hist_prefix = nice_scale_prefix(hist_f/f1)
        ax.set_ylabel(f'{hist_prefix}A')
    else:
        ax.set_ylabel(f'{hist_prefix}C/{ux}')
    

    ax.set_xlabel(labelx)  
    
    return fig
        
def marginal_plot(particle_group, key1='t', key2='p', bins=None, **kwargs):
    """
    Density plot and projections
    
    Example:
    
        marginal_plot(P, 't', 'energy', bins=200)   
    
    """
    
    if not bins:
        n = len(particle_group)
        bins = int(np.sqrt(n/4) )

    # Scale to nice units and get the factor, unit prefix
    x, f1, p1 = nice_array(particle_group[key1])
    y, f2, p2 = nice_array(particle_group[key2])

    w = particle_group['weight']
    
    u1 = particle_group.units(key1).unitSymbol
    u2 = particle_group.units(key2).unitSymbol
    ux = p1+u1
    uy = p2+u2
    
    labelx = f'{key1} ({ux})'
    labely = f'{key2} ({uy})'
    
    fig = plt.figure(**kwargs)
    
    gs = GridSpec(4,4)
    
    ax_joint =  fig.add_subplot(gs[1:4,0:3])
    ax_marg_x = fig.add_subplot(gs[0,0:3])
    ax_marg_y = fig.add_subplot(gs[1:4,3])
    #ax_info = fig.add_subplot(gs[0, 3:4])
    #ax_info.table(cellText=['a'])
    
    # Proper weighting
    ax_joint.hexbin(x, y, C=w, reduce_C_function=np.sum, gridsize=bins, cmap=CMAP0, vmin=1e-20)
    
    # Manual histogramming version
    #H, xedges, yedges = np.histogram2d(x, y, weights=w, bins=bins)
    #extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
    #ax_joint.imshow(H.T, cmap=cmap, vmin=1e-16, origin='lower', extent=extent, aspect='auto')
    
    
    
    # Top histogram
    # Old method:
    #dx = x.ptp()/bins
    #ax_marg_x.hist(x, weights=w/dx/f1, bins=bins, color='gray')
    hist, bin_edges = np.histogram(x, bins=bins, weights=w)
    hist_x = bin_edges[:-1] + np.diff(bin_edges) / 2
    hist_width =  np.diff(bin_edges)
    hist_y, hist_f, hist_prefix = nice_array(hist/hist_width)
    ax_marg_x.bar(hist_x, hist_y, hist_width, color='gray')
    # Special label for C/s = A
    if u1 == 's':
        _, hist_prefix = nice_scale_prefix(hist_f/f1)
        ax_marg_x.set_ylabel(f'{hist_prefix}A')
    else:
        ax_marg_x.set_ylabel(f'{hist_prefix}C/{ux}')
    
    
    # Side histogram
    # Old method:
    #dy = y.ptp()/bins
    #ax_marg_y.hist(y, orientation="horizontal", weights=w/dy, bins=bins, color='gray')
    hist, bin_edges = np.histogram(y, bins=bins, weights=w)
    hist_x = bin_edges[:-1] + np.diff(bin_edges) / 2
    hist_width =  np.diff(bin_edges)
    hist_y, hist_f, hist_prefix = nice_array(hist/hist_width)
    ax_marg_y.barh(hist_x, hist_y, hist_width, color='gray')
    ax_marg_y.set_xlabel(f'{hist_prefix}C/{uy}')
    
    # Turn off tick labels on marginals
    plt.setp(ax_marg_x.get_xticklabels(), visible=False)
    plt.setp(ax_marg_y.get_yticklabels(), visible=False)
    
    # Set labels on joint
    ax_joint.set_xlabel(labelx)
    ax_joint.set_ylabel(labely)

    return fig    
    
    
def density_and_slice_plot(particle_group, key1='t', key2='p', stat_keys=['norm_emit_x', 'norm_emit_y'], bins=100, n_slice=30):
    """
    Density plot and projections
    
    Example:
    
        marginal_plot(P, 't', 'energy', bins=200)   
    
    """

    # Scale to nice units and get the factor, unit prefix
    x, f1, p1 = nice_array(particle_group[key1])
    y, f2, p2 = nice_array(particle_group[key2])
    w = particle_group['weight']
    
    u1 = particle_group.units(key1).unitSymbol
    u2 = particle_group.units(key2).unitSymbol
    ux = p1+u1
    uy = p2+u2
    
    labelx = f'{key1} ({ux})'
    labely = f'{key2} ({uy})'
    
    fig, ax = plt.subplots()
    
    ax.set_xlabel(labelx)
    ax.set_ylabel(labely)
    
    # Proper weighting
    #ax_joint.hexbin(x, y, C=w, reduce_C_function=np.sum, gridsize=bins, cmap=cmap, vmin=1e-15)
    
    # Manual histogramming version
    H, xedges, yedges = np.histogram2d(x, y, weights=w, bins=bins)
    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
    ax.imshow(H.T, cmap=CMAP0, vmin=1e-16, origin='lower', extent=extent, aspect='auto')
    
    # Slice data
    slice_dat = slice_statistics(particle_group, n_slice=n_slice, slice_key=key1,
                            keys= stat_keys + ['ptp_'+key1, 'mean_'+key1, 'charge'])
    
    slice_dat['density'] = slice_dat['charge']/ slice_dat['ptp_'+key1]
    
    # 
    ax2 = ax.twinx()
    #ax2.set_ylim(0, 1e-6)
    x2 = slice_dat['mean_'+key1] / f1
    ulist = [particle_group.units(k).unitSymbol for k in stat_keys]
    
    max2 = max([slice_dat[k].ptp() for k in stat_keys])
    
    f3, p3 = nice_scale_prefix(max2)
    
    u2 = ulist[0]
    assert all([u==u2 for u in ulist] )
    u2 = p3 + u2
    for k in stat_keys:
        ax2.plot(x2, slice_dat[k]/f3, label=k)
    ax2.legend()     
    ax2.set_ylabel(f'({u2})')  
    ax2.set_ylim(bottom=0)
    
    
    # Add density
    y2 = slice_dat['density']
    y2 = y2 * max2/y2.max() / f3 /2
    ax2.fill_between(x2, 0, y2, color='black', alpha = 0.1)  
    
    
    
    
    
    
#-------------------------------------
#-------------------------------------
# Fields

def plot_fieldmesh_cylindrical_2d(fm,
                                  component=None,
                                  time=None,
                                  axes=None,
                                  aspect='auto',
                                  cmap=None,
                                  return_figure=False,
                                  **kwargs):
    """
    Plots a fieldmesh component
    
    
    """
    
    assert fm.geometry == 'cylindrical'
    
    if component is None:
        if fm.is_pure_magnetic:
            component='B'
        else:
            component='E'
    
    if not axes:
        fig, ax = plt.subplots(**kwargs)
        
        
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)        
        
    if not cmap:
        cmap = CMAP1       
    
   
    unit = fm.units(component)
    
    xmin, _, zmin = fm.mins
    xmax, _, zmax = fm.maxs
    
    # plt.imshow on [r, z] will put z on the horizontal axis. 
    extent = [zmin, zmax, xmin, xmax]
    
    xlabel = 'z (m)'
    ylabel = 'r (m)'

    
    dat = fm[component][:,0,:]
    dat = np.real_if_close(dat)
    
    dmin = dat.min()
    dmax = dat.max()
    
    ax.set_aspect(aspect)
    # Need to flip for image
    ax.imshow(np.flipud(dat), extent=extent, cmap=cmap, aspect=aspect)
    
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    
    # Add legend
    llabel = f'{component} ({unit.unitSymbol})'
    
    norm = matplotlib.colors.Normalize(vmin=dmin, vmax=dmax)
    fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap),
             cax=cax, orientation='vertical', label=llabel)     
    
    
    if return_figure:
        return fig


from pmd_beamphysics.units import e_charge
from pmd_beamphysics.interfaces.superfish import fish_complex_to_real_fields

import numpy as np
import subprocess
import os
   
def write_gpt(particle_group,           
               outfile,
               asci2gdf_bin=None,
               verbose=False): 

    """
    GPT uses a custom binary format GDF for particle data. This can be created with the
    asci2gdf utility as:
    asci2gdf -o particles.gdf particles.txt
    
    This routine makes ASCII particles, with column labels:
        'x', 'y', 'z', 'GBx', 'GBy', 'GBz', 't', 'q', 'nmacro'
    in SI units. 
    
    For now, only electrons are supported.

    """

    assert particle_group.species == 'electron' # TODO: add more species
    
    assert np.all(particle_group.weight > 0), 'ParticleGroup.weight must be > 0'
    
    q = -e_charge
    
    n = particle_group.n_particle
    gamma = particle_group.gamma
    
    dat = {        
        'x': particle_group.x,
        'y': particle_group.y,
        'z': particle_group.z,
        'GBx': gamma*particle_group.beta_x,
        'GBy': gamma*particle_group.beta_y,
        'GBz': gamma*particle_group.beta_z,
        't': particle_group.t,
        'q': np.full(n, q),
        'nmacro': particle_group.weight/e_charge}
    
    if hasattr(particle_group, 'id'):
        dat['ID'] = particle_group.id
    else:
        dat['ID'] = np.arange(1, particle_group['n_particle']+1)       
    
    
    header = ' '.join(list(dat))

    outdat = np.array([dat[k] for k in dat]).T
    
    if verbose:
        print(f'writing {n} particles to {outfile}')
    
    
    # Write ASCII
    np.savetxt(outfile, outdat, header=header, comments='', fmt = '%20.12e')
    
    if asci2gdf_bin:
        run_asci2gdf(outfile, asci2gdf_bin)
    else: 
        print(f'ASCII particles written. Convert to GDF using: asci2df -o particles.gdf {outfile}')
        

def run_asci2gdf(outfile, asci2gdf_bin, verbose=False):
    """
    Helper function to convert ASCII to GDF using asci2gdf
    """
    tempfile = outfile+'.txt'
    os.rename(outfile, tempfile)
    
    asci2gdf_bin = os.path.expandvars(asci2gdf_bin)
    assert os.path.exists(asci2gdf_bin), f'{asci2gdf_bin} does not exist'
    cmd = [asci2gdf_bin, '-o', outfile, tempfile]
    if verbose:
        print(' '.join(cmd))
    result = subprocess.run(cmd)
    
    # Cleanup
    os.remove(tempfile)    
   
    if verbose:
        print('Written GDF file:', outfile)
    
    return outfile
    
    
    
def write_gpt_fieldmesh(fm,           
               outfile,
               asci2gdf_bin=None,
               verbose=False): 
    """
    Writes a GPT fieldmap file from a FieldMesh object.
    
    Requires cylindrical geometry for now.
    """
    
    assert fm.geometry == 'cylindrical', f'Geometry: {fm.geometry} not implemented'
    
    assert fm.shape[1] == 1, 'Cylindrical symmetry required'
    
    dat = {}
    dat['R'], dat['Z'] = np.meshgrid(fm.coord_vec('r'), fm.coord_vec('z'), indexing='ij')
    
    keys = ['R', 'Z']
    if fm.is_static:
        if fm.is_pure_magnetic:
            keys = ['R', 'Z', 'Br', 'Bz']
            dat['Br'] = np.real(fm['Br'][:,0,:])
            dat['Bz'] = np.real(fm['Bz'][:,0,:])
        elif fm.is_pure_electric:
            keys = ['R', 'Z', 'Er', 'Ez']
            dat['Er'] = np.real(fm['Er'][:,0,:])
            dat['Ez'] = np.real(fm['Ez'][:,0,:])            
        else:
            raise ValueError('Mixed static field TODO')
            
    else:
        # Use internal Superfish routine 
        keys = ['R', 'Z', 'Er', 'Ez', 'Bphi']
        dat['Er'], dat['Ez'], dat['Bphi'], _ = fish_complex_to_real_fields(fm, verbose=verbose)            

        
    # Flatten dat     
    gptdata = np.array([dat[k].flatten() for k in keys]).T            
    
    # Write file. 
    # Hack to delete final newline
    # https://stackoverflow.com/questions/28492954/numpy-savetxt-stop-newline-on-final-line
    with open(outfile, 'w') as fout:
        NEWLINE_SIZE_IN_BYTES = 1 # 2 on Windows?
        np.savetxt(fout, gptdata, header=' '.join(keys), comments='')
        fout.seek(0, os.SEEK_END) # Go to the end of the file.
        # Go backwards one byte from the end of the file.
        fout.seek(fout.tell() - NEWLINE_SIZE_IN_BYTES, os.SEEK_SET)
        fout.truncate() # Truncate the file to this point.    
        
        
    if asci2gdf_bin:
        run_asci2gdf(outfile, asci2gdf_bin, verbose=verbose)
    elif verbose: 
        print(f'ASCII field data written. Convert to GDF using: asci2df -o field.gdf {outfile}')        
            
    return outfile    
    
    

        

    
        



import numpy as np

c_light = 299792458.




def parse_impact_particles(filePath, 
                           names=('x', 'GBx', 'y', 'GBy', 'z', 'GBz'),
                           skiprows=0):
    """
    Parse Impact-T input and output particle data.
    Typical filenames: 'partcl.data', 'fort.40', 'fort.50'.
    
    Note that partcl.data has the number of particles in the first line, so skiprows=1 should be used.
    
    Returns a structured numpy array
    
    Impact-T input/output particles distribions are ASCII files with columns:
    x (m)
    GBy = gamma*beta_x (dimensionless)
    y (m)
    GBy = gamma*beta_y (dimensionless)
    z (m)
    GBz = gamma*beta_z (dimensionless)
    
    Routine from lume-impact: 
        https://github.com/ChristopherMayes/lume-impact
    
    """
    
    dtype={'names': names,
           'formats': 6*[np.float]}
    pdat = np.loadtxt(filePath, skiprows=skiprows, dtype=dtype,
                     ndmin=1) # to make sure that 1 particle is parsed the same as many.

    return pdat    
    



def impact_particles_to_particle_data(tout, mc2=0, species=None, time=0, macrocharge=0, cathode_kinetic_energy_ref=None, verbose=False):
    """
    Convert impact particles to data for ParticleGroup
    
    particle_charge is the charge in units of |e|
    
    At the cathode, Impact-T translates z to t = z / (beta*c) for emission,
    where (beta*c) is the velocity calculated from kinetic energy:
        header['Bkenergy'] in eV.
    This is purely a conversion factor. 
    
    If cathode_kinetic_energy_ref is given, z will be parsed appropriately to t, and z will be set to 0. 
    
    Otherwise, particles will be set to the same time.
    
    """
    
    #mc2 = SPECIES_MASS[species]
    assert mc2 >0, 'mc2 must be specified'
    assert species, 'species must be specified'
       
    data = {}
    
    n_particle = len(tout['x'])
    
    data['x'] = tout['x']
    data['y'] = tout['y']
    #data['z'] = tout['z'] will be handled below

    data['px'] = tout['GBx']*mc2
    data['py'] = tout['GBy']*mc2
    data['pz'] = tout['GBz']*mc2
    
    # Handle z
    if cathode_kinetic_energy_ref:
        # Cathode start
        z = np.full(n_particle, 0.0)
        
        # Note that this is purely a conversion factor. 
        gamma = 1.0 + cathode_kinetic_energy_ref/mc2
        betac = np.sqrt(1-1/gamma**2)*c_light
        
        t = tout['z']/betac
        if verbose:
            print(f'Converting z to t according to cathode_kinetic_energy_ref = {cathode_kinetic_energy_ref} eV')
        
        
    else:
        # Free space start
        z = tout['z']
        t = np.full(n_particle, time)
    
    data['z'] = z
    data['t'] = t
    
    data['status'] = np.full(n_particle, 1)
    if macrocharge == 0:
        weight = 1/n_particle
    else:
        weight = abs(macrocharge)
    data['weight'] =  np.full(n_particle, weight) 
    
    data['species'] = species
    data['n_particle'] = n_particle
    return data



def write_impact(particle_group,
                outfile,
                cathode_kinetic_energy_ref=None,
                include_header=True,
                verbose=False):
    """
    Writes Impact-T style particles from particle_group type data.
    
    outfile should ultimately be named 'partcl.data' for Impact-T
    
    For now, the species must be electrons. 
    
    If cathode_kinetic_energy_ref is given, t will be used to compute z for cathode emission.
    
    If include_header, the number of particles will be written as the first line. Default is True. 
    
    Otherwise, particles must have the same time, and should be started in free space.
    
    A dict is returned with info about emission, for use in Impact-T
    
    """

    def vprint(*a, **k):
        if verbose:
            print(*a, **k)
    
    n_particle = particle_group.n_particle
    
    vprint(f'writing {n_particle} particles to {outfile}')
    
    
    mc2 = particle_group.mass
    
    # Dict output for use in Impact-T
    output = {'input_particle_file':outfile}
    output['Np'] = n_particle
    
    # Handle z
    if cathode_kinetic_energy_ref:
        # Cathode start
    
        vprint(f'Cathode start with cathode_kinetic_energy_ref = {cathode_kinetic_energy_ref} eV')
        
        # Impact-T conversion factor in eV
        output['Bkenergy'] = cathode_kinetic_energy_ref
        
        # Note that this is purely a conversion factor. 
        gamma = 1.0 + cathode_kinetic_energy_ref/mc2
        betac = np.sqrt(1-1/gamma**2)*c_light
        
        t = particle_group['t']
        
        # All t must be negative. 
        t_ptp = t.ptp()
        # Allow for some padding
        if t_ptp == 0:
            t_ptp = 1e-15  # s
        t_pad = t_ptp*1e-3 # 0.1% pad 
        t_shift = -t.max() -t_pad
        
        # Suggest an emission time
        output['Temission'] = t_ptp + 2*t_pad
        
        # This is the time that particles are shifted
        #output['Temission_shift'] = t_shift
        # Change actual initial time to this shift
        output['Tini'] = t_shift
        
        tout = t+t_shift
        
        output['Temission_mean'] = tout.mean()
        
        z = (tout)*betac       
        
        # pz
        pz = particle_group['pz']
        #check for zero pz
        assert np.all(pz > 0), 'pz must be positive'
        
        gamma_beta_z = pz/mc2
        
    else:
        # Free space start
        z = particle_group['z']
        
        t = np.unique(particle_group['t'])
        assert len(t) == 1, 'All particles must be a the same time'
        t = t[0]
        output['Tini'] = t
        output['Flagimg'] = 0 # Turn off Cathode start
        gamma_beta_z = particle_group['pz']/mc2
        
        vprint(f'Normal start with at time {t} s')
    
    # Form data table
    dat = np.array([
        particle_group['x'],
        particle_group['px']/mc2,
        particle_group['y'],
        particle_group['py']/mc2,
        z,
        gamma_beta_z
    ])
    
    # Save to ASCII
    if include_header:
        header=str(n_particle)
    else:
        header=''
        
    np.savetxt(outfile, dat.T, header=header, comments='')
    
    # Return info dict
    return output



from pmd_beamphysics.units import dimension, dimension_name, SI_symbol, pg_units, c_light

from pmd_beamphysics.interfaces.astra import write_astra
from pmd_beamphysics.interfaces.bmad import write_bmad
from pmd_beamphysics.interfaces.genesis import write_genesis4_distribution, genesis2_beam_data,  write_genesis2_beam_file
from pmd_beamphysics.interfaces.gpt import write_gpt
from pmd_beamphysics.interfaces.impact import write_impact
from pmd_beamphysics.interfaces.litrack import write_litrack
from pmd_beamphysics.interfaces.lucretia import write_lucretia
from pmd_beamphysics.interfaces.opal import write_opal
from pmd_beamphysics.interfaces.elegant import write_elegant

from pmd_beamphysics.plot import density_plot, marginal_plot

from pmd_beamphysics.readers import particle_array, particle_paths
from pmd_beamphysics.species import charge_of, mass_of

from pmd_beamphysics.statistics import norm_emit_calc, normalized_particle_coordinate, particle_amplitude, particle_twiss_dispersion, matched_particles
from pmd_beamphysics.writers import write_pmd_bunch, pmd_init

from h5py import File
import numpy as np
from copy import deepcopy
import os


#-----------------------------------------
# Classes


class ParticleGroup:
    """
    Particle Group class
    
    Initialized on on openPMD beamphysics particle group:
        h5 = open h5 handle, or str that is a file
        data = raw data
    
    The required bunch data is stored in .data with keys
        np.array: x, px, y, py, z, pz, t, status, weight
        str: species
    where:
        x, y, z are positions in units of [m]
        px, py, pz are momenta in units of [eV/c]
        t is time in [s]
        weight is the macro-charge weight in [C], used for all statistical calulations.
        species is a proper species name: 'electron', etc. 
        
    Optional data:
        np.array: id
    where:
        id is a list of unique integers that identify the particles. 
    
        
    Derived data can be computed as attributes:
        .gamma, .beta, .beta_x, .beta_y, .beta_z: relativistic factors [1].
        .r, .theta: cylidrical coordinates [m], [1]
        .pr, .ptheta: momenta in the radial and angular coordinate directions [eV/c]
        .Lz: angular momentum about the z axis [m*eV/c]
        .energy : total energy [eV]
        .kinetic_energy: total energy - mc^2 in [eV]. 
        .higher_order_energy: total energy with quadratic fit in z or t subtracted [eV]
        .p: total momentum in [eV/c]
        .mass: rest mass in [eV]
        .xp, .yp: Slopes x' = dx/dz = dpx/dpz and y' = dy/dz = dpy/dpz [1].
        
    Normalized transvere coordinates can also be calculated as attributes:
        .x_bar, .px_bar, .y_bar, .py_bar in [sqrt(m)]
        The normalization is automatically calculated from the covariance matrix. 
        See functions in .statistics for more advanced usage.
        
        Their cooresponding amplitudes are:
        .Jx, .Jy [m]
        where Jx = (x_bar^2 + px_bar^2 )/2, 
        The momenta are normalized by the mass, so that:
            <Jx> = norm_emit_x
        and similar for y. 
        
    Statistics of any of these are calculated with:
        .min(X)
        .max(X)
        .ptp(X)
        .avg(X)
        .std(X)
        .cov(X, Y, ...)
        .histogramdd(X, Y, ..., bins=10, range=None)
        with a string X as the name any of the properties above.
        
    Useful beam physics quantities are given as attributes:
        .norm_emit_x
        .norm_emit_y
        .norm_emit_4d
        .higher_order_energy_spread
        .average_current
        
    Twiss parameters, including dispersion, for the 'x' or 'y' plane:
        .twiss(plane='x', fraction=0.95, p0C=None)
    For convenience, plane='xy' will calculate twiss for both planes.
    
    Twiss matched particles, using a simple linear transformation:
        .twiss_match(self, beta=None, alpha=None, plane='x', p0c=None, inplace=False)
    
                
    The weight is required and must sum to > 0. The sum of the weights is:
        .charge
    This can also be set:
        .charge = 1.234 # C, will rescale the .weight array
            
    All attributes can be accessed with brackets:
        [key]
    Additional keys are allowed for convenience:
        ['min_prop']   will return  .min('prop')
        ['max_prop']   will return  .max('prop')
        ['ptp_prop']   will return  .ptp('prop')
        ['mean_prop']  will return  .avg('prop')
        ['sigma_prop'] will return  .std('prop')
        ['cov_prop1__prop2'] will return .cov('prop1', 'prop2')[0,1]
        
    Units for all attributes can be accessed by:
        .units(key)
    
    Particles are often stored at the same time (i.e. from a t-based code), 
    or with the same z position (i.e. from an s-based code.)
    Routines: 
        drift_to_z(z0)
        drift_to_t(t0)
    help to convert these. If no argument is given, particles will be drifted to the mean.
        
    
    """
    def __init__(self, h5=None, data=None):
    
    
        if h5 and data:
            # TODO:
            # Allow merging or changing some array with extra data
            raise NotImplementedError('Cannot init on both h5 and data')
    
        if h5:
            # Allow filename
            if isinstance(h5, str):
                fname = os.path.expandvars(h5)
                assert os.path.exists(fname), f'File does not exist: {fname}'
  
                with File(fname, 'r') as hh5:
                    pp = particle_paths(hh5)
                    assert len(pp) == 1, f'Number of particle paths in {h5}: {len(pp)}'
                    data = load_bunch_data(hh5[pp[0]])

            else:
                # Try dict
                data = load_bunch_data(h5)
        else:
            # Fill out data. Exclude species.
            data = full_data(data)
            species = list(set(data['species']))
            
            # Allow for empty data (len=0). Otherwise, check species.
            if len(species) >= 1:
                assert len(species) == 1, f'mixed species are not allowed: {species}'
                data['species'] = species[0]
            
        self._settable_array_keys = ['x', 'px', 'y', 'py', 'z', 'pz', 't', 'status', 'weight']
        # Optional data
        for k in ['id']:
            if k in data:
                self._settable_array_keys.append(k)  
            
        self._settable_scalar_keys = ['species']
        self._settable_keys =  self._settable_array_keys + self._settable_scalar_keys                       
        # Internal data. Only allow settable keys
        self._data = {k:data[k] for k in self._settable_keys}
        
    #-------------------------------------------------
    # Access to intrinsic coordinates
    @property
    def x(self):
        """
        x coordinate in [m]
        """
        return self._data['x']
    @x.setter
    def x(self, val):
        self._data['x'] = full_array(len(self), val)    
            
    @property
    def y(self):
        """
        y coordinate in [m]
        """
        return self._data['y']
    @y.setter
    def y(self, val):
        self._data['y'] = full_array(len(self), val)              
       
    @property
    def z(self):
        """
        z coordinate in [m]
        """
        return self._data['z']
    @z.setter
    def z(self, val):
        self._data['z'] = full_array(len(self), val)      
        
    @property
    def px(self):
        """
        px coordinate in [eV/c]
        """
        return self._data['px']
    @px.setter
    def px(self, val):
        self._data['px'] = full_array(len(self), val)      
        
    @property
    def py(self):
        """
        py coordinate in [eV/c]
        """
        return self._data['py']
    @py.setter
    def py(self, val):
        self._data['py'] = full_array(len(self), val)    
        
    @property
    def pz(self):
        """
        pz coordinate in [eV/c]
        """
        return self._data['pz']
    @pz.setter
    def pz(self, val):
        self._data['pz'] = full_array(len(self), val)    
        
    @property
    def t(self):
        """
        t coordinate in [s]
        """
        return self._data['t']
    @t.setter
    def t(self, val):
        self._data['t'] = full_array(len(self), val)       
        
    @property
    def status(self):
        """
        status coordinate in [1]
        """
        return self._data['status']
    @status.setter
    def status(self, val):
        self._data['status'] = full_array(len(self), val)  

    @property
    def weight(self):
        """
        weight coordinate in [C]
        """
        return self._data['weight']
    @weight.setter
    def weight(self, val):
        self._data['weight'] = full_array(len(self), val)            
        
    @property
    def id(self):
        """
        id integer 
        """
        if 'id' not in self._data:
            self.assign_id()      
        
        return self._data['id']
    @id.setter
    def id(self, val):
        self._data['id'] = full_array(len(self), val)            
    
    
    @property
    def species(self):
        """
        species string
        """
        return self._data['species']
    @species.setter
    def species(self, val):
        self._data['species'] = val
        
    @property
    def data(self):
        """
        Internal data dict
        """
        return self._data        
    
    #-------------------------------------------------
    # Derived data
            
    def assign_id(self):
        """
        Assigns unique ids, integers from 1 to n_particle
        
        """
        if 'id' not in self._settable_array_keys: 
            self._settable_array_keys.append('id')
        self.id = np.arange(1, self['n_particle']+1)             
    
    @property
    def n_particle(self):
        """Total number of particles. Same as len """
        return len(self)
    
    @property
    def n_alive(self):
        """Number of alive particles, defined by status == 1"""
        return len(np.where(self.status==1)[0])
    
    @property
    def n_dead(self):
        """Number of alive particles, defined by status != 1"""
        return self.n_particle - self.n_alive
    
        
    def units(self, key):
        """Returns the units of any key"""
        return pg_units(key)
        
    @property
    def mass(self):
        """Rest mass in eV"""
        return mass_of(self.species)

    @property
    def species_charge(self):
        """Species charge in C"""
        return charge_of(self.species)
    
    @property
    def charge(self):
        """Total charge in C"""
        return np.sum(self.weight)
    @charge.setter
    def charge(self, val):
        """Rescale weight array so that it sum to this value"""
        assert val >0, 'charge must be >0. This is used to weight the particles.'
        self.weight *= val/self.charge
        
    
    # Relativistic properties
    @property
    def p(self):
        """Total momemtum in eV/c"""
        return np.sqrt(self.px**2 + self.py**2 + self.pz**2) 
    @property
    def energy(self):
        """Total energy in eV"""
        return np.sqrt(self.px**2 + self.py**2 + self.pz**2 + self.mass**2) 
    @property
    def kinetic_energy(self):
        """Kinetic energy in eV"""
        return self.energy - self.mass
    
    # Slopes. Note that these are relative to pz
    @property
    def xp(self):
        """x slope px/pz (dimensionless)"""
        return self.px/self.pz  
    @property
    def yp(self):
        """y slope py/pz (dimensionless)"""
        return self.py/self.pz    
    
    @property
    def higher_order_energy(self):
        """
        Fits a quadratic (order=2) to the Energy vs. time, and returns the energy with this subtracted. 
        """ 
        return self.higher_order_energy_calc(order=2)

    @property
    def higher_order_energy_spread(self):
        """
        Legacy syntax to compute the standard deviation of higher_order_energy.
        """
        return self.std('higher_order_energy')
     
    def higher_order_energy_calc(self, order=2):
        """
        Fits a polynmial with order `order` to the Energy vs. time, , and returns the energy with this subtracted. 
        """
        #order=2
        if self.std('z') < 1e-12:
            # must be at a screen. Use t
            t = self.t
        else:
            # All particles at the same time. Use z to calc t
            t = self.z/c_light
        energy = self.energy
        
        best_fit_coeffs = np.polynomial.polynomial.polyfit(t, energy, order)
        best_fit = np.polynomial.polynomial.polyval(t, best_fit_coeffs)
        return energy - best_fit
    
    # Polar coordinates. Note that these are centered at x=0, y=0, and not an averge center. 
    @property
    def r(self):
        """Radius in the xy plane: r = sqrt(x^2 + y^2) in m"""
        return np.hypot(self.x, self.y)
    @property    
    def theta(self):
        """Angle in xy plane: theta = arctan2(y, x) in radians"""
        return np.arctan2(self.y, self.x)
    @property
    def pr(self):
        """
        Momentum in the radial direction in eV/c
        r_hat = cos(theta) xhat + sin(theta) yhat
        pr = p dot r_hat
        """
        theta = self.theta
        return self.px * np.cos(theta)  + self.py * np.sin(theta)   
    @property    
    def ptheta(self):
        """     
        Momentum in the polar theta direction. 
        theta_hat = -sin(theta) xhat + cos(theta) yhat
        ptheta = p dot theta_hat
        Note that Lz = r*ptheta
        """
        theta = self.theta
        return -self.px * np.sin(theta)  + self.py * np.cos(theta)  
    @property    
    def Lz(self):
        """
        Angular momentum around the z axis in m*eV/c
        Lz = x * py - y * px
        """
        return self.x*self.py - self.y*self.px
    
    
    # Relativistic quantities
    @property
    def gamma(self):
        """Relativistic gamma"""
        return self.energy/self.mass
    
    @gamma.setter
    def gamma(self, val):
        beta_x = self.beta_x
        beta_y = self.beta_y
        beta_z = self.beta_z
        beta = self.beta
        gamma_new = full_array(len(self), val)
        energy_new = gamma_new * self.mass
        beta_new = np.sqrt(gamma_new**2 - 1)/gamma_new
        self._data['px'] = energy_new * beta_new * beta_x / beta
        self._data['py'] = energy_new * beta_new * beta_y / beta
        self._data['pz'] = energy_new * beta_new * beta_z / beta
        
    @property
    def beta(self):
        """Relativistic beta"""
        return self.p/self.energy
    @property
    def beta_x(self):
        """Relativistic beta, x component"""
        return self.px/self.energy
    
    @beta_x.setter
    def beta_x(self, val):
        self._data['px'] = full_array(len(self), val)*self.energy    
        
    @property
    def beta_y(self):
        """Relativistic beta, y component"""
        return self.py/self.energy
    
    @beta_y.setter
    def beta_y(self, val):
        self._data['py'] = full_array(len(self), val)*self.energy    
        
    @property
    def beta_z(self):
        """Relativistic beta, z component"""
        return self.pz/self.energy
    
    @beta_z.setter
    def beta_z(self, val):
        self._data['pz'] = full_array(len(self), val)*self.energy
    
    # Normalized coordinates for x and y
    @property 
    def x_bar(self):
        """Normalized x in units of sqrt(m)"""
        return normalized_particle_coordinate(self, 'x')
    @property     
    def px_bar(self):
        """Normalized px in units of sqrt(m)"""
        return normalized_particle_coordinate(self, 'px')    
    @property
    def Jx(self):
        """Normalized amplitude J in the x-px plane"""
        return particle_amplitude(self, 'x')
    
    @property 
    def y_bar(self):
        """Normalized y in units of sqrt(m)"""
        return normalized_particle_coordinate(self, 'y')
    @property     
    def py_bar(self):
        """Normalized py in units of sqrt(m)"""
        return normalized_particle_coordinate(self, 'py')
    @property
    def Jy(self):
        """Normalized amplitude J in the y-py plane"""
        return particle_amplitude(self, 'y')    
    
    def delta(self, key):
        """Attribute (array) relative to its mean"""
        return getattr(self, key) - self.avg(key)
      
    
    # Statistical property functions
    
    def min(self, key):
        """Minimum of any key"""
        return np.min(getattr(self, key))
    def max(self, key):
        """Maximum of any key"""
        return np.max(getattr(self, key)) 
    def ptp(self, key):
        """Peak-to-Peak = max - min of any key"""
        return np.ptp(getattr(self, key))     
        
    def avg(self, key):
        """Statistical average"""
        dat = getattr(self, key) # equivalent to self.key for accessing properties above
        if np.isscalar(dat): 
            return dat
        return np.average(dat, weights=self.weight)
    def std(self, key):
        """Standard deviation (actually sample)"""
        dat = getattr(self, key)
        if np.isscalar(dat):
            return 0
        avg_dat = self.avg(key)
        return np.sqrt(np.average( (dat - avg_dat)**2, weights=self.weight))
    def cov(self, *keys):
        """
        Covariance matrix from any properties
    
        Example: 
        P = ParticleGroup(h5)
        P.cov('x', 'px', 'y', 'py')
    
        """
        dats = np.array([ getattr(self, key) for key in keys ])
        return np.cov(dats, aweights=self.weight)
    
    def histogramdd(self, *keys, bins=10, range=None):
        """
        Wrapper for numpy.histogramdd, but accepts property names as keys.
        
        Computes the multidimensional histogram of keys. Internally uses weights. 
        
        Example:
            P.histogramdd('x', 'y', bins=50)
        Returns:
            np.array with shape 50x50, edge list 
        
        """
        H, edges = np.histogramdd(np.array([self[k] for k in list(keys)]).T, weights=self.weight, bins=bins, range=range)
        
        return H, edges
    
    
    
    
    # Beam statistics
    @property
    def norm_emit_x(self):
        """Normalized emittance in the x plane"""
        return norm_emit_calc(self, planes=['x'])
    @property
    def norm_emit_y(self):       
        """Normalized emittance in the x plane"""
        return norm_emit_calc(self, planes=['y'])
    @property
    def norm_emit_4d(self):       
        """Normalized emittance in the xy planes (4D)"""
        return norm_emit_calc(self, planes=['x', 'y'])    
    
    def twiss(self, plane='x', fraction=1, p0c=None):
        """
        Returns Twiss and Dispersion dict.
        
        plane can be:
            'x', 'y', 'xy'
        
        Optionally a fraction of the particles, based on amplitiude, can be specified.
        """
        d = {}
        for p in plane:
            d.update(particle_twiss_dispersion(self, plane=p, fraction=fraction, p0c=p0c))
        return d
   
    def twiss_match(self, beta=None, alpha=None, plane='x', p0c=None, inplace=False):
        """
        Returns a ParticleGroup with requested Twiss parameters.
        
        See: statistics.matched_particles
        """
        
        return matched_particles(self, beta=beta, alpha=alpha, plane=plane, inplace=inplace)
        
        
    
    
    
    @property
    def average_current(self):
        """
        Simple average current in A: charge / dt, with dt =  (max_t - min_t)
        If particles are in t coordinates, will try dt = (max_z - min_z)*c_light*beta_z
        """
        dt = self.t.ptp()  # ptp 'peak to peak' is max - min
        if dt == 0:
            # must be in t coordinates. Calc with 
            dt = self.z.ptp() / (self.avg('beta_z')*c_light)
        return self.charge / dt
    
    def __getitem__(self, key):
        """
        Returns a property or statistical quantity that can be computed:
        P['x'] returns the x array
        P['sigmx_x'] returns the std(x) scalar
        P['norm_emit_x'] returns the norm_emit_x scalar
        
        Parts can also be given. Example: P[0:10] returns a new ParticleGroup with the first 10 elements.
        """
        
        # Allow for non-string operations: 
        if not isinstance(key, str):
            return particle_parts(self, key)
    
        if key.startswith('cov_'):
            subkeys = key[4:].split('__')
            assert len(subkeys) == 2, f'Too many properties in covariance request: {key}'
            return self.cov(*subkeys)[0,1]
        elif key.startswith('delta_'):
            return self.delta(key[6:])
        elif key.startswith('sigma_'):
            return self.std(key[6:])
        elif key.startswith('mean_'):
            return self.avg(key[5:])
        elif key.startswith('min_'):
            return self.min(key[4:])
        elif key.startswith('max_'):
            return self.max(key[4:])     
        elif key.startswith('ptp_'):
            return self.ptp(key[4:])         
        
        else:
            return getattr(self, key) 
    
    def where(self, x):
        return self[np.where(x)]
    
    # TODO: should the user be allowed to do this?
    #def __setitem__(self, key, value):    
    #    assert key in self._settable_keyes, 'Error: you cannot set:'+str(key)
    #    
    #    if key in self._settable_array_keys:
    #        assert len(value) == self.n_particle
    #        self.__dict__[key] = value
    #    elif key == 
    #        print()
     
    # Simple 'tracking'     
    def drift(self, delta_t):
        """
        Drifts particles by time delta_t
        """
        self.x = self.x + self.beta_x * c_light * delta_t
        self.y = self.y + self.beta_y * c_light * delta_t
        self.z = self.z + self.beta_z * c_light * delta_t
        self.t = self.t + delta_t
    
    def drift_to_z(self, z=None):

        if not z:
            z = self.avg('z')
        dt = (z - self.z) / (self.beta_z * c_light)
        self.drift(dt)
        # Fix z to be exactly this value
        self.z = np.full(self.n_particle, z)
        
        
    def drift_to_t(self, t=None):
        """
        Drifts all particles to the same t
        
        If no z is given, particles will be drifted to the average t
        """
        if not t:
            t = self.avg('t')
        dt = t - self.t
        self.drift(dt)
        # Fix t to be exactly this value
        self.t = np.full(self.n_particle, t)
    
    # Writers
    def write_astra(self, filePath, verbose=False):
        write_astra(self, filePath, verbose=verbose)
        
    def write_bmad(self, filePath, p0c=None, t_ref=0, verbose=False):
        write_bmad(self, filePath, p0c=p0c, t_ref=t_ref, verbose=verbose)        

    def write_elegant(self, filePath, verbose=False):
        write_elegant(self, filePath, verbose=verbose)            
        
    def write_genesis2_beam_file(self, filePath, n_slice=None, verbose=False):
        # Get beam columns 
        beam_columns = genesis2_beam_data(self, n_slice=n_slice)
        # Actually write the file
        write_genesis2_beam_file(filePath, beam_columns, verbose=verbose)  
        
    def write_genesis4_distribution(self, filePath, verbose=False):
        write_genesis4_distribution(self, filePath, verbose=verbose)
        
    def write_gpt(self, filePath, asci2gdf_bin=None, verbose=False):
        write_gpt(self, filePath, asci2gdf_bin=asci2gdf_bin, verbose=verbose)    
    
    def write_impact(self, filePath, cathode_kinetic_energy_ref=None, include_header=True, verbose=False):
        return write_impact(self, filePath, cathode_kinetic_energy_ref=cathode_kinetic_energy_ref,
                            include_header=include_header, verbose=verbose)          
        
    def write_litrack(self, filePath, p0c=None, verbose=False):        
        return write_litrack(self, outfile=filePath, p0c=p0c, verbose=verbose)      
        
    def write_lucretia(self, filePath, ele_name='BEGINNING', t_ref=0, stop_ix=None, verbose=False):       
        return write_lucretia(self, filePath, ele_name=ele_name, t_ref=t_ref, stop_ix=stop_ix)
        
    def write_opal(self, filePath, verbose=False, dist_type='emitted'):
        return write_opal(self, filePath, verbose=verbose, dist_type=dist_type)
    
        
    # openPMD    
    def write(self, h5, name=None):
        """
        Writes to an open h5 handle, or new file if h5 is a str.
        
        """
        if isinstance(h5, str):
            fname = os.path.expandvars(h5)
            g = File(fname, 'w')
            pmd_init(g, basePath='/', particlesPath='.' )
        else:
            g = h5
    
        write_pmd_bunch(g, self, name=name)        
        
        
    # Plot
    # TODO: more general plotting
    def plot(self, key1='x', key2=None, bins=None, return_figure=False, **kwargs):
        """
        1d or 2d density plot. 
        """
        
        if not key2:
            fig = density_plot(self, key=key1, bins=bins, **kwargs)
        else:
            fig = marginal_plot(self, key1=key1, key2=key2, bins=bins, **kwargs)
        
        if return_figure:
            return fig
        
    # New constructors
    def split(self, n_chunks = 100, key='z'):
        return split_particles(self, n_chunks=n_chunks, key=key)
    
    def copy(self):
        """Returns a deep copy"""
        return deepcopy(self)    
    
    
    # Operator overloading    
    
    # Resample
    def resample(self, n):
        """
        Resamples n particles.
        """
        return resample(self, n)
    
    # Internal sorting
    def _sort(self, key):
        """Sorts internal arrays by key"""
        ixlist = np.argsort(self[key])
        for k in self._settable_array_keys:
            self._data[k] = self[k][ixlist]    
        
    # Operator overloading    
    def __add__(self, other):
        """
        Overloads the + operator to join particle groups.
        Simply calls join_particle_groups
        """
        return join_particle_groups(self, other)
    
    # 
    def __contains__(self, item):
        """Checks internal data"""
        return True if item in self._data else False    
    
    def __len__(self):
        return len(self[self._settable_array_keys[0]])
    
    def __str__(self):
        s = f'ParticleGroup with {self.n_particle} particles with total charge {self.charge} C'
        return s

    def __repr__(self):
        memloc = hex(id(self))
        return f'<ParticleGroup with {self.n_particle} particles at {memloc}>'
                   


#-----------------------------------------
# helper functions for ParticleGroup class
    
    
def single_particle(x=0,
                   px=0,
                   y=0,
                   py=0,
                   z=0,
                   pz=0,
                   t=0,
                   weight=1,
                   status=1,
                   species='electron'):
    """
    Convenience function to make ParticleGroup with a single particle.
    
    Units:
        x, y, z: m
        px, py, pz: eV/c
        t: s
        weight: C
        status=1 => live particle
        
    """
    data = dict(x=x, px=px, y=y, py=py, z=z, pz=pz, t=t, weight=weight, status=status, species=species)
    return ParticleGroup(data=data)
    
def load_bunch_data(h5):
    """
    Load particles into structured numpy array.
    """
    n = len(h5['position/x'])
    
    attrs = dict(h5.attrs)
    data = {}
    data['species'] = attrs['speciesType'].decode('utf-8') # String
    n_particle = int(attrs['numParticles'])
    data['total_charge'] = attrs['totalCharge']*attrs['chargeUnitSI']
    
    for key in ['x', 'px', 'y', 'py', 'z', 'pz', 't']:
        data[key] = particle_array(h5, key)
        
    if 'particleStatus' in h5:
        data['status'] = particle_array(h5, 'particleStatus')
    else:
        data['status'] = np.full(n_particle, 1)
    
    # Make sure weight is populated
    if 'weight' in h5:
        weight = particle_array(h5, 'weight')
        if len(weight) == 1:
            weight = np.full(n_particle, weight[0])
    else:
        weight = np.full(n_particle, data['total_charge']/n_particle)
    data['weight'] = weight
    
    # id should be a unique integer, no units
    # optional
    if 'id' in h5:
        data['id'] = h5['id'][:]
        
    return data


def full_array(n, val):
    """
    Casts a value into a full array of length n
    """
    if np.isscalar(val):
        return np.full(n, val)
    n_here = len(val)
    
    if n_here == 1:
        return np.full(n, val[0])
    elif n_here != n:
        raise ValueError(f'Length mismatch: len(val)={n_here}, but requested n={n}')
    # Cast to array
    return np.array(val)    
    
    

def full_data(data, exclude=None):
    """
    Expands keyed data into np arrays, assuring that the lengths of all items are the same. 
    
    Allows for some keys to be scalars or length 1, and fills them out with np.full.
    
    
    """
    
    full_data = {}
    scalars = {}
    for k, v in data.items():
        if np.isscalar(v):
            scalars[k] = v
        elif len(v) == 1:
            scalars[k] = v[0]
        else:
            # must be array
            full_data[k] = np.array(v)
    
    # Check for single particle
    if len(full_data) == 0:
        return {k:np.array([v]) for k, v in scalars.items()}
            
    # Array data should all have the same length
    nlist = [len(v) for _, v in full_data.items()]
    assert len(set(nlist)) == 1, f'arrays must have the same length. Found len: { {k:len(v) for k, v in full_data.items()} }'
    
    for k, v in scalars.items():
        full_data[k] = np.full(nlist[0], v)
    
    return full_data


def split_particles(particle_group, n_chunks = 100, key='z'):
    """
    Splits a particle group into even chunks. Returns a list of particle groups. 
    
    Useful for creating slice statistics. 
    
    """
    
    # Sorting
    zlist = getattr(particle_group, key) 
    iz = np.argsort(zlist)

    # Split particles into chunks
    plist = []
    for chunk in np.array_split(iz, n_chunks):
        # Prepare data
        data = {}
        #keys = ['x', 'px', 'y', 'py', 'z', 'pz', 't', 'status', 'weight'] 
        for k in particle_group._settable_array_keys:
            data[k] = getattr(particle_group, k)[chunk]
        # These should be scalars
        data['species'] = particle_group.species
        
        # New object
        p = ParticleGroup(data=data)
        plist.append(p)
        
    return plist

def resample(particle_group, n):
    """
    Resamples a ParticleGroup randomly.
    
    Returns a new ParticleGroup instance.
    
    Note that this only works if the weights are the same
    
    """
    n_old = particle_group.n_particle
    assert n <= n_old, 'Cannot supersample'
    assert len(set(particle_group.weight)) == 1, 'non-unique weights for resampling.'
    ixlist = np.random.choice(n_old, n, replace=False)
    data = {}
    for key in particle_group._settable_array_keys:
        data[key] = particle_group[key][ixlist]
    data['species'] = particle_group['species']
    data['weight'] *= n_old/n # Need to re-weight
    
    return ParticleGroup(data=data)



def particle_parts(particle_group, x):
    """
    Gets parts of a ParticleGroup object. Returns a new ParticleGroup
    """
    data = {}
    for k in particle_group._settable_array_keys:
        data[k] = particle_group[k][x]

    for k in particle_group._settable_scalar_keys:
        data[k] = particle_group[k]

    return ParticleGroup(data=data)
           
    
    


    
def join_particle_groups(*particle_groups):
    """
    Join particle groups. 
    
    This simply concatenates the internal particle arrays.
    
    Species must be the same
    """
    species = [pg['species'] for pg in particle_groups]
    #return species 

    species0 = species[0]
    assert all([spe == species0 for spe in species]) , 'species must be the same to join'
    
    data = {}
    for key in particle_groups[0]._settable_array_keys:
        data[key] = np.hstack([pg[key] for pg in particle_groups ])
    
    data['species'] = species0
    data['n_particle'] = np.sum( [pg['n_particle'] for pg in particle_groups]) 
    
    return ParticleGroup(data=data)    
    
    

    


        







from scipy.interpolate import UnivariateSpline
import numpy as np

def expand_1d_static_fieldmap(z0, fz0):
    """
    Expands 1D static fieldmap z, fz into r, z using splines.
    
    Cylindrically symmetric geometry. 
    
    This is valid for both electric and magnetic fields. 
    
    Returns functions: 
        fr(r, z), fz(r, z)
    
    """
    
    # Make spline and derivatives
    S = UnivariateSpline(z0, fz0, k=5, s=0)
    Sp = S.derivative()
    Sp2 = S.derivative(2)
    Sp3 = S.derivative(3)
    
    def fz(r, z):   
        return S(z) - r**2 * Sp2(z)/4
        
    def fr(r, z):   
        return -r*Sp(z)/2 + r**3 * Sp3(z)/16   
    
    return fr, fz



def expand_1d_dynamic_fieldmap(z, Ez0, frequency=0):
    """
    Expands 1D dynamic elecric fieldmap z, Ez into r, z using splines.
    
    Cylindrically symmetric geometry. 
    
    Fields oscillate as:
        Er, Ez ~ sin(wt)  <=> ~  cos(wt)
        Btheta ~ cos(wt)      ~ -sin(wt)
    This is the Superfish convention. 
        
    
    Returns real functions: 
        Er(r, z), Ez(r, z), Btheta(r, z)
            
    """
    
    omega = 2*np.pi*frequency
    c_light=299792458.
    ooc2 = (omega/c_light)**2 
    
    # Make spline and derivatives
    S = UnivariateSpline(z, Ez0, k=5, s=0)
    Sp = S.derivative()
    Sp2 = S.derivative(2)
    Sp3 = S.derivative(3)
    
    def Ez(r, z):   
        return S(z) - r**2 /4 * (Sp2(z) + (omega/c_light)**2 * S(z))  # sin(omega*t)
        
    def Er(r, z):   
        return -r/2 * Sp(z) + r**3/16 * (Sp3(z) + (omega/c_light)**2 * Sp(z)) # sin(omega*t)
    
    def Btheta(r,z):
        return (r/2 * S(z) - r**3 /16 * (Sp2(z) + (omega/c_light)**2 *S(z) ))*omega/c_light**2. # cos(omega*t)
    
    
    return Er, Ez, Btheta


import numpy as np



def norm_emit_calc(particle_group, planes=['x']):
    """
    
    2d, 4d, 6d normalized emittance calc
    
    planes = ['x', 'y'] is the 4d emittance
    
    planes = ['x', 'y', 'z'] is the 6d emittance
    
    Momenta for each plane are takes as p+plane, e.g. 'px' for plane='x'
    
    The normalization factor is (1/mc)^n_planes, so that the units are meters^n_planes
    
    """
    
    dim = len(planes)
    vars = []
    for k in planes:
        vars.append(k)
        vars.append('p'+k)
    
    S = particle_group.cov(*vars)
    
    mc2 = particle_group.mass
    
    norm_emit = np.sqrt(np.linalg.det(S)) / mc2**dim
    
    return norm_emit


def twiss_calc(sigma_mat2):
    """
    Calculate Twiss parameters from the 2D sigma matrix (covariance matrix):
    sigma_mat = <x,x>   <x, p>
                <p, x>  <p, p>

    This is a simple calculation. Makes no assumptions about units. 
    
    alpha = -<x, p>/emit
    beta  =  <x, x>/emit
    gamma =  <p, p>/emit
    emit = det(sigma_mat)
    
    """
    assert sigma_mat2.shape == (2,2), f'Bad shape: {sigma_mat2.shape}. This should be (2,2)' # safety check
    twiss={}
    emit = np.sqrt(np.linalg.det(sigma_mat2)) 
    twiss['alpha'] = -sigma_mat2[0,1]/emit
    twiss['beta']  = sigma_mat2[0,0]/emit
    twiss['gamma'] = sigma_mat2[1,1]/emit
    twiss['emit']  = emit
    
    return twiss



def twiss_match(x, p, beta0=1, alpha0=0, beta1=1, alpha1=0):
    """
    Simple Twiss matching. 
    
    Takes positions x and momenta p, and transforms them according to 
    initial Twiss parameters:
        beta0, alpha0 
    into final  Twiss parameters:
        beta1, alpha1
        
    This is simply the matrix ransformation:    
        xnew  = (   sqrt(beta1/beta0)                  0                 ) . ( x )
        pnew    (  (alpha0-alpha1)/sqrt(beta0*beta1)   sqrt(beta0/beta1) )   ( p ) 
        

    Returns new x, p
    
    """
    m11 = np.sqrt(beta1/beta0)
    m21 = (alpha0-alpha1)/np.sqrt(beta0*beta1)
    
    xnew = x * m11
    pnew = x * m21 + p / m11
    
    return xnew, pnew

def matched_particles(particle_group, beta=None, alpha=None, plane='x', p0c=None, inplace=False):
    """
    Perfoms simple Twiss 'matching' by applying a linear transformation to
        x, px if plane == 'x', or x, py if plane == 'y'
    
    Returns a new ParticleGroup
    
    If inplace, a copy will not be made, and changes will be done in place. 
    
    """
    
    assert plane in ('x', 'y'), f'Invalid plane: {plane}'
    
    if inplace:
        P = particle_group
    else:
        P = particle_group.copy()
    
    if not p0c:
        p0c = P['mean_p']
    

    # Use Bmad-style coordinates.
    # Get plane. 
    if plane == 'x':
        x = P.x
        p = P.px/p0c
    else:
        x = P.y
        p = P.py/p0c
        
    # Get current Twiss
    tx = twiss_calc(np.cov(x, p, aweights=P.weight))
    
    # If not specified, just fill in the current value.
    if alpha is None:
        alpha = tx['alpha']
    if beta is None:
        beta = tx['beta']
    
    # New coordinates
    xnew, pnew = twiss_match(x, p, beta0=tx['beta'], alpha0=tx['alpha'], beta1=beta, alpha1=alpha)
    
    # Set
    if plane == 'x':
        P.x = xnew
        P.px = pnew*p0c
    else:
        P.y = xnew
        P.py = pnew*p0c
        
    return P





def twiss_dispersion_calc(sigma3):
    """
    Twiss and Dispersion calculation from a 3x3 sigma (covariance) matrix from particles
    x, p, delta
    
    Formulas from: 
        https://uspas.fnal.gov/materials/19Knoxville/g-2/creation-and-analysis-of-beam-distributions.html
    
    Returns a dict with:
        alpha
        beta
        gamma
        emit
        eta      
        etap
    
    """
    
    # Collect terms
    
    delta2 = sigma3[2,2]
    xd = sigma3[0,2]
    pd = sigma3[1,2]
    
    eb =  sigma3[0,0] - xd**2 / delta2
    eg =  sigma3[1,1] - pd**2 / delta2
    ea = -sigma3[0,1] + xd * pd / delta2
    
    emit = np.sqrt(eb*eg - ea**2)
    
    # Form the output dict
    d = {}

    d['alpha'] = ea/emit
    d['beta']  = eb/emit
    d['gamma'] = eg/emit
    d['emit'] = emit
    d['eta']  = xd/delta2
    d['etap'] = pd/delta2

    return d


def particle_twiss_dispersion(particle_group, plane='x', fraction=1, p0c=None):
    """
    Twiss and Dispersion calc for a ParticleGroup. 
    
    Plane muse be:
        'x' or 'y'
    
    p0c is the reference momentum. If not give, the mean p will be used.
    
    Returns the same output dict as twiss_dispersion_calc, but with keys suffixed with the plane, i.e.:
     
        alpha_x
        beta_x
        gamma_x
        emit_x
        eta_x      
        etap_x
        norm_emit_x

    """
    
    assert plane in ['x', 'y']
    
    P = particle_group # convenience
    
    if fraction < 1:
        P = P[np.argsort(P[f'J{plane}'])][0:int(fraction*len(P))]
    
    if not p0c:
        p0c = P['mean_p'] 
    
    x = P[plane]
    xp = P['p'+plane]/p0c
    delta = P['p']/P['mean_p'] # - 1
    
    # Form covariance matrix
    np.cov([x, xp, delta], aweights=P.weight)
    
    # Actual calc
    twiss = twiss_dispersion_calc(np.cov([x, xp, delta]))
    
    # Add norm
    twiss['norm_emit'] = twiss['emit'] * P['mean_p']/P.mass
    
    # Add suffix
    out = {}
    for k in twiss:
        out[k+f'_{plane}'] = twiss[k]
    
    return out










# Linear Normal Form in 1 phase space plane.
# TODO: more advanced analysis e.g. Forest or Wolski or Sagan and Rubin or Ehrlichman. 

def A_mat_calc(beta, alpha, inverse=False):
    """
    Returns the 1D normal form matrix from twiss parameters beta and alpha
    
        A =   sqrt(beta)         0 
             -alpha/sqrt(beta)   1/sqrt(beta) 
    
    If inverse, the inverse will be returned:

        A^-1 =  1/sqrt(beta)     0 
                alpha/sqrt(beta) sqrt(beta)     
    
    This corresponds to the linear normal form decomposition:
    
        M = A . Rot(theta) . A^-1
    
    with a clockwise rotation matrix:
    
        Rot(theta) =  cos(theta) sin(theta)
                     -sin(theta) cos(theta)
    
    In the Bmad manual, G_q (Bmad) = A (here) in the Linear Optics chapter.
    
    A^-1 can be used to form normalized coordinates: 
        x_bar, px_bar   = A^-1 . (x, px)
    
    """
    a11 = np.sqrt(beta)  
    a22 = 1/a11
    a21 = -alpha/a11
    
    if inverse:
        return np.array([
                        [a22, 0],
                        [-a21, a11]])        
    else:
        return np.array([
                    [a11, 0],
                    [a21, a22]])
    

def amplitude_calc(x, p, beta=1, alpha=0):
    """
    Simple amplitude calculation of position and momentum coordinates
    relative to twiss beta and alpha. 

    J = (gamma x^2 + 2 alpha x p + beta p^2)/2
        
      = (x_bar^2 + px_bar^2)/ 2
      
    where gamma = (1+alpha^2)/beta
    
    """
    return (1+alpha**2)/beta/2 * x**2 + alpha*x*p + beta/2*p**2


def particle_amplitude(particle_group, plane='x', twiss=None, mass_normalize=True):
    """
    Returns the normalized amplitude array from a ParticleGroup for a given plane.
    
    Plane should be:
        'x' for the x, px plane
        'y' for the y, py plane 
    Other planes will work, but please check that the units make sense. 
        
    If mass_normalize (default=True), the momentum will be divided by the mass, so that the units are sqrt(m).        
    
    See: normalized_particle_coordinate
    """
    x = particle_group[plane]
    key2 = 'p'+plane
    
    if mass_normalize:
        # Note: do not do /=, because this will replace the ParticleGroup's internal array!
        p = particle_group[key2]/particle_group.mass
    else:
        p = particle_group[key2] 
        
    # User could supply twiss    
    if not twiss:
        sigma_mat2 = np.cov(x, p, aweights=particle_group.weight)
        twiss = twiss_calc(sigma_mat2)           
    
    J = amplitude_calc(x, p, beta=twiss['beta'], alpha=twiss['alpha'])
    
    return J 
    
    
    
def normalized_particle_coordinate(particle_group, key, twiss=None, mass_normalize=True):
    """
    Returns a single normalized coordinate array from a ParticleGroup
    
    Position or momentum is determined by the key. 
    If the key starts with 'p', it is a momentum, else it is a position,
    and the
    
    Intended use is for key to be one of:
        x, px, y py
        
    and the corresponding normalized coordinates are named with suffix _bar, i.e.:
        x_bar, px_bar, y_bar, py_bar
    
    If mass_normalize (default=True), the momentum will be divided by the mass, so that the units are sqrt(m).
    
    These are related to action-angle coordinates
        J: amplitude
        phi: phase
        
        x_bar =  sqrt(2 J) cos(phi)
        px_bar = sqrt(2 J) sin(phi)
        
    So therefore:
        J = (x_bar^2 + px_bar^2)/2
        phi = arctan(px_bar/x_bar)
    and:        
        <J> = norm_emit_x
     
     Note that the center may need to be subtracted in this case.
     
    """
    
    # Parse key for position or momentum coordinate
    if key.startswith('p'):
        momentum = True
        key1 = key[1:]
        key2 = key
        
    else:
        momentum = False
        key1 = key
        key2 = 'p'+key
            
    x = particle_group[key1]
    
    if mass_normalize:
        # Note: do not do /=, because this will replace the ParticleGroup's internal array!
        p = particle_group[key2]/particle_group.mass
    else:
        p = particle_group[key2]
    
     
    # User could supply twiss    
    if not twiss:
        sigma_mat2 = np.cov(x, p, aweights=particle_group.weight)
        twiss = twiss_calc(sigma_mat2)        
    
    A_inv = A_mat_calc(twiss['beta'], twiss['alpha'], inverse=True)

    if momentum:
        return A_inv[1,0]*x + A_inv[1,1]*p
    else:
        return A_inv[0,0]*x    
    
    
# ---------------
# Other utilities
    
def slice_statistics(particle_group,  keys=['mean_z'], n_slice=40, slice_key='z'):
    """
    Slices a particle group into n slices and returns statistics from each sliced defined in keys. 
    
    These statistics should be scalar floats for now.
    
    Any key can be used to slice on. 
    
    """
    sdat = {}
    for k in keys:
        sdat[k] = np.empty(n_slice)
    for i, pg in enumerate(particle_group.split(n_slice, key=slice_key)):
        for k in keys:
            sdat[k][i] = pg[k]
            
    return sdat






import scipy.constants
mu_0 = scipy.constants.mu_0

import numpy as np








# ------------------
# FieldMesh write T7

def write_fish_t7(fm, filePath, fmt='%10.8e', verbose=False):
    """
    Writes a T7 file from FISH t7data dict. 
    
    Input:
        fm: FieldMesh object
        filePath: requested filePath to write
    
    See:
        superfish.parsers.parse_fish_t7
    """
    
    
    assert fm.geometry == 'cylindrical', 'TODO: cartesian.'
    assert fm.frequency != 0, 'Frequency must be non-zero.'
    
    rmin, _, zmin = fm.mins
    rmax, _, zmax = fm.maxs
    nr, _, nz = list(fm.shape)
    
    
    # Collect these. Units are cm, MHz
    xmin = zmin*100
    xmax = zmax*100
    nx   = nz
    ymin = rmin*100
    ymax = rmax*100
    ny   = nr
    freq = fm.frequency*1e-6
        
    # Get complex fields (helper function)
    Er, Ez, Btheta, _ =  fish_complex_to_real_fields(fm, verbose=verbose)

    # Scale to Superfish units
    Er *= 1e-6 # V/m -> MV/m
    Ez *= 1e-6
    Hphi = Btheta/mu_0  # convert to H field, and in Superfish phase convention. 
    E = np.hypot(Er, Ez)
 
    # Write T7 ASCII
    header = f"""{xmin} {xmax} {nx-1}
{freq}
{ymin} {ymax} {ny-1}"""
    
    # Unroll the arrays
    dat = np.array([field.reshape(nx*ny).T for field in [Ez, Er, E, Hphi]]).T
    
    np.savetxt(filePath, dat, header=header, comments='',  fmt = fmt)
    
    if verbose:
        print(f"Superfish T7 file '{filePath}' written for Fish problem.")    
    
    return filePath


def fish_complex_to_real_fields(fm, verbose=False):
    """
    Internal routine for .interfaces.superfish.write_fish_t7
    
    Takes complex Ez, Er, Btheta and determines the complex angle to rotate Ez onto the real axis.
    
    Returns rotated real fields:
        np.real(Er), np.real(Ez), -np.imag(Btheta), rot_angle
    
    """
    # Get complex fields
    Ez = fm.Ez[:,0,:]
    Er = fm.Er[:,0,:]
    Btheta = fm.Btheta[:,0,:]
    
    # Get imaginary argument (angle)
    iangle = np.unique(np.mod(np.angle(Ez), np.pi))

    # Make sure there is only one
    assert len(iangle) == 1, print(iangle)
    iangle = iangle[0]
    
    if iangle != 0:
        rot = np.exp(-1j*iangle)
        if verbose:
            print(f'Rotating complex field by {-iangle}')
    else:
        rot = 1
    
    Er = np.real(Er*rot)
    Ez = np.real(Ez*rot)
    Btheta = -np.imag(Btheta*rot)

    return Er, Ez, Btheta, -iangle
    

def write_poisson_t7(fm, filePath, fmt='%10.8e', verbose=False):
    """
    Writes a T7 file from POISSON t7data dict. 
    
    
    See:
        superfish.parsers.parse_poisson_t7
    """
    
    
    assert fm.geometry == 'cylindrical', 'TODO: cartesian.'
    assert fm.is_static, 'Static fields are required for Poisson T7'
    
    rmin, _, zmin = fm.mins
    rmax, _, zmax = fm.maxs
    nr, _, nz = list(fm.shape)
    
    
    # Collect these. Units are cm
    # Note: different from FISH!
    ymin = zmin*100
    ymax = zmax*100
    ny   = nz
    xmin = rmin*100
    xmax = rmax*100
    nx   = nr
    
    # Write T7 ASCII
    header = f"""{xmin} {xmax} {nx-1}
{ymin} {ymax} {ny-1}"""
    
    
    if fm.is_pure_electric:
        kr = 'Er'
        kz = 'Ez'
        ftype = 'electric'
        factor = 1
        
    else:
        kr = 'Br'
        kz = 'Bz'
        ftype = 'magnetic'
        factor = 1e4 # T->G
    fr = np.real(fm[kr][:,0,:])*factor
    fz = np.real(fm[kz][:,0,:])*factor
    
    # Unroll the arrays
    dat = np.array([field.reshape(nx*ny, order='F').T for field in [fr, fz]]).T
    
    np.savetxt(filePath, dat, header=header, comments='',  fmt = fmt)
    
    if verbose:
        print(f"Superfish T7 file '{filePath}' written for {ftype} Poisson problem.")
    
    return filePath


def write_superfish_t7(fm, filePath, fmt='%10.8e', verbose=False):
    """
    Writes a Superfish T7 file. This is a simple wrapper for:
        write_fish_t7
        write_poisson_t7
    If .is_statice, a Poisson file is written. Otherwise a Fish file is written.
    
    """
    if fm.is_static:
        return write_poisson_t7(fm, filePath, verbose=verbose)
    else:
        return write_fish_t7(fm, filePath, verbose=verbose)



# ------------------
# Parsers for ASCII T7


def read_superfish_t7(t7file,
                      type='electric',
                      geometry='cylindrical'):
    """
    Parses a T7 file written by Posson/Superfish.
    
    Fish or Poisson T7 are automatically detected according to the second line.
    
    For Poisson problems, the type must be specified.
    
    Superfish fields oscillate as:
        Er, Ez ~ cos(wt)
        Hphi   ~ -sin(wt)
      
    For complex fields oscillating as e^-iwt
    
        Re(Ex*e^-iwt)   ~ cos
        Re(-iB*e^-iwt) ~ -sin        
    and therefore B = -i * mu_0 * H_phi is the complex magnetic field in Tesla
    

    Parameters:
    ----------
    t7file: str
    
    type: str, optional
        For Poisson files, required to be 'electric' or 'magnetic'. 
        Not used for Fish files
    geometry: str, optional
        field geometry, currently required to be the default: 'cylindrical'
    
    Returns:
    -------
    fieldmesh_data: dict of dicts:
        attrs
        components
        
        
    A FieldMesh object is instantiated from this as:
        FieldMesh(data=fieldmesh_data)
    
    """
    
    # ASCII parsing
    
    # Read header
    # zmin(cm), zmax(cm), nx-1
    # freq(MHz)
    # ymin(cm), ymax(cm), ny-1
    with open(t7file, 'r') as f:
        line1 = f.readline().split()
        line2 = f.readline().split()
        line3 = f.readline().split()
     
    if len(line2) == 1:
        problem = 'fish'
    else:
        problem = 'poisson'
    
    components = {}
    if problem=='fish':
        # zmin(cm), zmax(cm), nz-1
        # freq(MHz)
        # rmin(cm), rmax(cm), nr-1
        # 4 columns of data: Ez(MV/m), Er(MV/m), E(MV/m), Hphi(A/m)
        
        # FISH problem   
        zmin, zmax, nz =  float(line1[0])*1e-2, float(line1[1])*1e-2, int(line1[2])+1
        frequency = float(line2[0])*1e6 # MHz -> Hz
        rmin, rmax, nr =  float(line3[0])*1e-2, float(line3[1])*1e-2, int(line3[2])+1  
        
        # Read and reshape
        dat = np.loadtxt(t7file, skiprows=3)
        #labels=['Ez', 'Er', 'E', 'Hphi']
        dat = dat.reshape(nr, 1, nz, 4)
        
        components['electricField/z'] = dat[:,:,:,0].astype(np.complex) * 1e6 # MV/m -> V/m
        components['electricField/r'] = dat[:,:,:,1].astype(np.complex) * 1e6 # MV/m -> V/m
        components['magneticField/theta'] = dat[:,:,:,3]  * -1j*mu_0 # A/m -> T
        
    else:
        # rmin(cm), rmax(cm), nx-1    # r in cylindrical geometry
        # zmin(cm), zmax(cm), ny-1    # z in cylindrical geometry
        
        # POISSON problem
        rmin, rmax, nr =  float(line1[0])*1e-2, float(line1[1])*1e-2, int(line1[2])+1
        zmin, zmax, nz =  float(line2[0])*1e-2, float(line2[1])*1e-2, int(line2[2])+1        
        frequency=0
        
        # The structure here is different 
        dat = np.loadtxt(t7file, skiprows=2)
        dat = dat.reshape(nz, 1, nr, 2)
        
        # type must be specified
        if type == 'electric':
            components['electricField/r'] = dat[:,:,:,0].T # V/m
            components['electricField/z'] = dat[:,:,:,1].T # V/m
        elif type == 'magnetic':
            components['magneticField/r'] = dat[:,:,:,0].T*1e-4 # G -> T
            components['magneticField/z'] = dat[:,:,:,1].T*1e-4 # G -> T
    
    
    dz = (zmax-zmin)/(nz-1)
    dr = (rmax)/(nr-1)
    
    # Attributes
    attrs = {}
    attrs['eleAnchorPt'] = 'beginning'
    attrs['gridGeometry'] = 'cylindrical'
    attrs['axisLabels'] = ('r', 'theta', 'z')
    attrs['gridLowerBound'] = (0, 1, 0)
    attrs['gridSize'] = (nr, 1, nz)        
    attrs['gridSpacing'] = (dr, 0, dz)    
    
    # Set requested zmin
    attrs['gridOriginOffset'] = (0, 0, zmin) 
    
    attrs['fundamentalFrequency'] = frequency
    attrs['RFphase'] = 0
    if frequency == 0:
        attrs['harmonic'] = 0
    else:
        attrs['harmonic'] = 1
    
    
    return dict(attrs=attrs, components=components)


